/**
 * @file iwl-io.c
 *
 * @details
 * This file contains code to read, write, clear etc. operations
 * for UNDI driver.
 *
 * @copyright
 * Copyright (c) 2017 Dell Inc. All rights reserved. This
 * software and associated documentation (if any) is furnished
 * under a license and may only be used or copied in accordance
 * with the terms of the license. Except as permitted by such
 * license, no part of this software or documentation may be
 * reproduced, stored in a retrieval system, or transmitted in
 * any form or by any means without the express written consent
 * of Dell Inc.
 *
 * Dell elects to use only the BSD license for any software where
 * a choice of BSD/GPLv2 license versions are made available with
 * the language indicating that BSD/GPLv2 or any later version may
 * be used, or where a choice of which version of the BSD/GPLv2 is
 * applied is unspecified.
 */

/******************************************************************************
 *
 * This file is provided under a dual BSD/GPLv2 license.  When using or
 * redistributing this file, you may do so under either license.
 *
 * GPL LICENSE SUMMARY
 *
 * Copyright(c) 2012 - 2014 Intel Corporation. All rights reserved.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of version 2 of the GNU General Public License as
 * published by the Free Software Foundation.
 *
 * This program is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110,
 * USA
 *
 * The full GNU General Public License is included in this distribution
 * in the file called COPYING.
 *
 * Contact Information:
 *  Intel Linux Wireless <ilw@linux.intel.com>
 * Intel Corporation, 5200 N.E. Elam Young Parkway, Hillsboro, OR 97124-6497
 *
 * BSD LICENSE
 *
 * Copyright(c) 2012 - 2014 Intel Corporation. All rights reserved.
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 *
 *  * Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 *  * Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in
 *    the documentation and/or other materials provided with the
 *    distribution.
 *  * Neither the name Intel Corporation nor the names of its
 *    contributors may be used to endorse or promote products derived
 *    from this software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
 * "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
 * LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
 * A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
 * OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 * SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
 * LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
 * DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
 * THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
 * OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 *
 *****************************************************************************/

#include <Uefi.h>
#include "iwl-io.h"
#include "internal.h"
#include "iwl-prph.h"

#define IWL_POLL_INTERVAL 10    /* microseconds */


UINT32
IwlReadPrph (
    IN IWL_TRANS_CONTEXT *IwlTransContext,
    IN UINT32 Ofs
    )
{
    UINT32 ReadVal = 0x5A5A5A5A;

    if(IwlTransContext != NULL)
    {
        if (IwlTransPcieGrabNicAccess(IwlTransContext, FALSE ))
        {
            ReadVal = IwlTransPcieReadPrph(IwlTransContext,Ofs);
            IwlTransPcieReleaseNicAccess(IwlTransContext);
        }
    }
    return (ReadVal);
}

VOID
IwlWritePrph (
    IN IWL_TRANS_CONTEXT *IwlTransContext,
    IN u32 Ofs,
    IN u32 Val
    )
{

    if (IwlTransPcieGrabNicAccess(IwlTransContext, FALSE)) {

        IwlTransPcieWritePrph(IwlTransContext, Ofs, Val);

        IwlTransPcieReleaseNicAccess(IwlTransContext);
    }
}

VOID
IwlSetBitsPrph (
    IN IWL_TRANS_CONTEXT *IwlTransContext,
    IN UINT32 Ofs,
    IN UINT32 Mask
    )
{
    if(IwlTransPcieGrabNicAccess(IwlTransContext,FALSE)) {

        IwlTransPcieWritePrph(IwlTransContext, Ofs,(IwlTransPcieReadPrph(IwlTransContext, Ofs) | Mask));

        IwlTransPcieReleaseNicAccess(IwlTransContext);
    }
}

VOID
IwlClearBitsPrph (
    IN IWL_TRANS_CONTEXT *IwlTransContext,
    IN u32 Ofs,
    IN u32 Mask
    )
{
    u32 Val;

    if (IwlTransPcieGrabNicAccess(IwlTransContext, FALSE)) {

        Val = IwlTransPcieReadPrph(IwlTransContext, Ofs);

        IwlTransPcieWritePrph(IwlTransContext, Ofs, (Val & ~Mask));

        IwlTransPcieReleaseNicAccess(IwlTransContext);
    }
}

INT32
IwlPollDirectBit (
    IN IWL_TRANS_CONTEXT *IwlTransContext,
    IN u32 Addr,
    IN u32 Mask,
    IN INTN Timeout
    )
{
    INT32 t = 0;

    do {
        if ((IwlReadDirect32(IwlTransContext, Addr) & Mask) == Mask)
            return t;

        gBS->Stall(IWL_POLL_INTERVAL);//Mandatory delay added as per linux implementation
        t += IWL_POLL_INTERVAL;
    } while (t < Timeout);

    return -ETIMEDOUT;
}

void Iwl_Force_Nmi(struct iwl_trans_context *IwlTransContext)
{
	if(IwlTransContext != NULL)
	{
        if (IwlTransContext->IwlConfig->Device_Family != IWL_DEVICE_FAMILY_8000) {
            IwlWritePrph(IwlTransContext, DEVICE_SET_NMI_REG,
                         DEVICE_SET_NMI_VAL_DRV);
            IwlWritePrph(IwlTransContext, DEVICE_SET_NMI_REG,
                         DEVICE_SET_NMI_VAL_HW);
        }
        else
        {
            IwlWritePrph(IwlTransContext, DEVICE_SET_NMI_8000_REG,
                         DEVICE_SET_NMI_8000_VAL);
            IwlWritePrph(IwlTransContext, DEVICE_SET_NMI_REG,
                         DEVICE_SET_NMI_VAL_DRV);
        }
	}
}
