/**
 * @file iwl-op-mode.h
 *
 * @details
 * This header file contains function prototypes, structures related
 * to different types of operation modes, like queue full, move start,
 * NIC error etc. for Intel UNDI driver.
 *
 * @copyright
 * Copyright (c) 2017 Dell Inc. All rights reserved. This
 * software and associated documentation (if any) is furnished
 * under a license and may only be used or copied in accordance
 * with the terms of the license. Except as permitted by such
 * license, no part of this software or documentation may be
 * reproduced, stored in a retrieval system, or transmitted in
 * any form or by any means without the express written consent
 * of Dell Inc.
 */

/******************************************************************************
 *
 * This file is provided under a dual BSD/GPLv2 license.  When using or
 * redistributing this file, you may do so under either license.
 *
 * GPL LICENSE SUMMARY
 *
 * Copyright(c) 2007 - 2014 Intel Corporation. All rights reserved.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of version 2 of the GNU General Public License as
 * published by the Free Software Foundation.
 *
 * This program is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110,
 * USA
 *
 * The full GNU General Public License is included in this distribution
 * in the file called COPYING.
 *
 * Contact Information:
 *  Intel Linux Wireless <ilw@linux.intel.com>
 * Intel Corporation, 5200 N.E. Elam Young Parkway, Hillsboro, OR 97124-6497
 *
 * BSD LICENSE
 *
 * Copyright(c) 2005 - 2014 Intel Corporation. All rights reserved.
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 *
 *  * Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 *  * Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in
 *    the documentation and/or other materials provided with the
 *    distribution.
 *  * Neither the name Intel Corporation nor the names of its
 *    contributors may be used to endorse or promote products derived
 *    from this software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
 * "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
 * LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
 * A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
 * OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 * SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
 * LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
 * DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
 * THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
 * OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 *
 *****************************************************************************/

#ifndef IWLOPMODE_H_
#define IWLOPMODE_H_

#include "fw-api-rs.h"
/******************************************************************/
/***
*			                TYPEDEFS
*/
/******************************************************************/
typedef struct iwl_op_mode IWL_OP_MODE;
typedef struct iwl_cfg IWL_CONFIG;
typedef struct iwl_fw IWL_FW;
typedef struct iwl_trans IWL_TRANS;
typedef struct iwlwifi_opmode_table Opmode_Table;
typedef struct iwl_rx_cmd_buffer Iwl_Rx_Cmd_Buffer;
typedef struct iwl_device_cmd Iwl_Device_Cmd;

/**
 * struct iwl_op_mode_ops - op_mode specific operations
 *
 * The op_mode exports its ops so that external components can start it and
 * interact with it. The driver layer typically calls the start and stop
 * handlers, the transport layer calls the others.
 *
 * All the handlers MUST be implemented
 *
 * @start: start the op_mode. The transport layer is already allocated.
 *	May sleep
 * @stop: stop the op_mode. Must free all the memory allocated.
 *	May sleep
 * @rx: Rx notification to the op_mode. rxb is the Rx buffer itself. Cmd is the
 *	HCMD this Rx responds to.
 *	This callback may sleep, it is called from a threaded IRQ handler.
 * @queue_full: notifies that a HW queue is full.
 *	Must be atomic and called with BH disabled.
 * @queue_not_full: notifies that a HW queue is not full any more.
 *	Must be atomic and called with BH disabled.
 * @hw_rf_kill:notifies of a change in the HW rf kill switch. True means that
 *	the radio is killed. Return %true if the device should be stopped by
 *	the transport immediately after the call. May sleep.
 * @free_skb: allows the transport layer to free skbs that haven't been
 *	reclaimed by the op_mode. This can happen when the driver is freed and
 *	there are Tx packets pending in the transport layer.
 *	Must be atomic
 * @nic_error: error notification. Must be atomic and must be called with BH
 *	disabled.
 * @cmd_queue_full: Called when the command queue gets full. Must be atomic and
 *	called with BH disabled.
 * @nic_config: configure NIC, called before firmware is started.
 *	May sleep
 * @wimax_active: invoked when WiMax becomes active. May sleep
 * @enter_d0i3: configure the fw to enter d0i3. May sleep.
 * @exit_d0i3: configure the fw to exit d0i3. May sleep.
 */

typedef struct iwl_op_mode_ops {

    IWL_OP_MODE *(*start)(
        IN IWL_TRANS *Trans,
        IN CONST IWL_CONFIG *Cfg,
        IN CONST IWL_FW *Fw
        );

    VOID (*stop)(
        IN IWL_OP_MODE *Op_Mode
        );
    int (*rx)(
        struct iwl_op_mode *op_mode,
        struct iwl_rx_cmd_buffer *rxb,
        struct iwl_device_cmd *cmd,
        BOOLEAN page_stolen
        );
    VOID (*queue_full)(
        IWL_OP_MODE *Op_Mode,
        INT32 Queue
        );
    VOID (*queue_not_full)(
        IWL_OP_MODE *Op_Mode,
        INT32 Queue
        );
    BOOLEAN (*hw_rf_kill)(
        IWL_OP_MODE *Op_Mode,
        BOOLEAN State
        );
    VOID (*nic_error)(
        IWL_OP_MODE *Op_Mode
        );
    VOID (*cmd_queue_full)(
        IWL_OP_MODE *Op_Mode
        );
    VOID (*nic_config)(
        IWL_OP_MODE *Op_Mode
        );
    IWL_OP_MODE* (*getContext)(
        );
    VOID (*Lq_Cmd_Enqueue)(
            struct iwl_lq_cmd *lq,
            INTN index
        );
    VOID (*wimax_active)(
        IWL_OP_MODE *Op_Mode
        );

}IWL_OP_MODE_OPS;

/**
 * struct iwl_op_mode - operational mode
 * @ops: pointer to its own ops
 *
 * This holds an implementation of the mac80211 / fw API.
 */
typedef struct iwl_op_mode {
    IWL_OP_MODE_OPS *Ops;
    VOID *Op_Mode_Specific;
}IWL_OP_MODE;


/**********************************************************************************/
/***
 *                              FUNCTIONS
 */
/**********************************************************************************/

STATIC
VOID
iwl_op_mode_queue_full (
    IN struct iwl_op_mode *op_mode,
    IN INT32 queue
    )
{
    op_mode->Ops->queue_full(op_mode, queue);
}
STATIC
VOID
iwl_op_mode_nic_error (
    IN struct iwl_op_mode *op_mode
    )
{
    op_mode->Ops->nic_error(op_mode);
}

STATIC
VOID
iwl_op_mode_cmd_queue_full (
    IN struct iwl_op_mode *op_mode)
{
    op_mode->Ops->cmd_queue_full(op_mode);
}

STATIC
IWL_OP_MODE *
Iwl_Op_Mode_Mvm_Start (
    IN IWL_TRANS *Trans,
    IN CONST IWL_CONFIG *Cfg,
    IN CONST IWL_FW *Fw
    );
STATIC
VOID
Iwl_Op_Mode_Mvm_Stop (
    IN IWL_OP_MODE *Op_Mode
    );

STATIC
INT32
Iwl_Mvm_Rx_Dispatch (
    struct iwl_op_mode *op_mode,
    struct iwl_rx_cmd_buffer *rxb,
    struct iwl_device_cmd *cmd,
    BOOLEAN page_stolen
    );

STATIC
VOID
Iwl_Mvm_Stop_Sw_Queue (
    IN IWL_OP_MODE *Op_Mode,
    IN INT32 Queue
    );

STATIC
VOID
Iwl_Mvm_Wake_Sw_Queue (
    IN IWL_OP_MODE *Op_Mode,
    INT32 Queue
    );

STATIC
BOOLEAN
Iwl_Mvm_Set_Hw_Rfkill_State (
    IWL_OP_MODE *Op_Mode,
    BOOLEAN State
    );

STATIC
VOID
Iwl_Mvm_Nic_Error (
    IN IWL_OP_MODE *Op_Mode
    );
STATIC
VOID
Iwl_Mvm_Cmd_Queue_Full (
    IN IWL_OP_MODE *Op_Mode
    );
STATIC
VOID
Iwl_Mvm_Nic_Config (
    IN IWL_OP_MODE *Op_Mode
    );
IWL_OP_MODE*
Iwl_Op_Mode_Mvm_GetContext (
    );
STATIC
INT32
Iwl_Mvm_Enter_D0i3 (
    IN IWL_OP_MODE *Op_Mode
    );
STATIC
INT32
Iwl_Mvm_Exit_D0i3 (
    IN IWL_OP_MODE *Op_Mode
    );

IWL_OP_MODE *_Op_Mode_Driver_Start(IWL_FW *iwl_appln_fw);

#endif
