/**
 * @file internal.h
 *
 * @details
 * This header file contains function prototypes, defines, etc.
 * for transmit and receive operation for Intel UNDI driver.
 *
 * @copyright
 * Copyright (c) 2017 Dell Inc. All rights reserved. This
 * software and associated documentation (if any) is furnished
 * under a license and may only be used or copied in accordance
 * with the terms of the license. Except as permitted by such
 * license, no part of this software or documentation may be
 * reproduced, stored in a retrieval system, or transmitted in
 * any form or by any means without the express written consent
 * of Dell Inc.
 */


/******************************************************************************
 *
 * Copyright(c) 2003 - 2014 Intel Corporation. All rights reserved.
 *
 * Portions of this file are derived from the ipw3945 project, as well
 * as portions of the ieee80211 subsystem header files.
 *
 * This program is free software; you can redistribute it and/or modify it
 * under the terms of version 2 of the GNU General Public License as
 * published by the Free Software Foundation.
 *
 * This program is distributed in the hope that it will be useful, but WITHOUT
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
 * FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License for
 * more details.
 *
 * You should have received a copy of the GNU General Public License along with
 * this program; if not, write to the Free Software Foundation, Inc.,
 * 51 Franklin Street, Fifth Floor, Boston, MA 02110, USA
 *
 * The full GNU General Public License is included in this distribution in the
 * file called LICENSE.
 *
 * Contact Information:
 *  Intel Linux Wireless <ilw@linux.intel.com>
 * Intel Corporation, 5200 N.E. Elam Young Parkway, Hillsboro, OR 97124-6497
 *
 *****************************************************************************/

#ifndef PCIEINTERNAL_H_
#define PCIEINTERNAL_H_

#include "iwl-io.h"
#include "iwl-fh.h"
#include "iwl-csr.h"
#include <Library/BaseMemoryLib.h>
#include <Library/UefiBootServicesTableLib.h>



//****************************Typedefs***********************************/

typedef struct s_undi_device_instance UNDI32_DEV;
typedef struct iwl_cfg IWL_CONFIG;
typedef struct _EFI_DRIVER_BINDING_PROTOCOL EFI_DRIVER_BINDING_PROTOCOL;
typedef struct Iwl_Host_Cmd IWL_HOST_CMD;
typedef UINTN size_t;
typedef struct iwl_tx_skb IWL_TX_SKB;

//****************************Macros***********************************/

#define CPU1_CPU2_SEPARATOR_SECTION 0xFFFFCCCC
#define PAGING_SEPARATOR_SECTION    0xAAAABBBB
#define RX_QUEUE_SIZE 256
#define RX_FREE_BUFFERS 64
#define BITS_PER_LONG 32


#define IWL_MAX_HW_QUEUES       32
#define MAX_NO_RECLAIM_CMDS     6

#define INTEL_7260_DEVICE_ID_1  0x08b2
#define INTEL_7260_DEVICE_ID_2  0x08b1
#define INTEL_3160_DEVICE_ID    0x08b3

#define PAGE_SIZE 4096

struct iwl_rx_mem_buffer {
    UINT64 page_dma;
    UINT64 *cpu_addr;
    VOID *Mapping;
    struct list_head list;
};

struct isr_statistics {
    u32 hw;
    u32 sw;
    u32 err_code;
    u32 sch;
    u32 alive;
    u32 rfkill;
    u32 ctkill;
    u32 wakeup;
    u32 rx;
    u32 tx;
    u32 unhandled;
};

/**
 * struct iwl_rxq - Rx queue
 * @bd: driver's pointer to buffer of receive buffer descriptors (rbd)
 * @bd_dma: bus address of buffer of receive buffer descriptors (rbd)
 * @pool:
 * @queue:
 * @read: Shared index to newest available Rx buffer
 * @write: Shared index to oldest written Rx packet
 * @free_count: Number of pre-allocated buffers in rx_free
 * @write_actual:
 * @rx_free: list of free SKBs for use
 * @rx_used: List of Rx buffers with no SKB
 * @need_update: flag to indicate we need to update read/write index
 * @rb_stts: driver's pointer to receive buffer status
 * @rb_stts_dma: bus address of receive buffer status
 * @lock:
 *
 * NOTE:  rx_free and rx_used are used as a FIFO for iwl_rx_mem_buffers
 */

struct iwl_rxq {
    UINT64 *bd;
    UINT64 bd_dma;
    struct iwl_rx_mem_buffer pool[RX_QUEUE_SIZE + RX_FREE_BUFFERS];
    struct iwl_rx_mem_buffer *queue[RX_QUEUE_SIZE];
    u32 read;
    u32 write;
    u32 free_count;
    u32 write_actual;
    struct list_head rx_free;
    struct list_head rx_used;
    int need_update;
    struct iwl_rb_status *rb_stts;
    UINT64 rb_stts_dma;
    VOID *Mapping_bd;
    VOID *Mapping_rb;
};

struct iwl_dma_ptr {
    dma_addr_t dma;
    void *addr;
    size_t size;
    void *mapping;
};

/**
 * iwl_queue_inc_wrap - increment queue index, wrap back to beginning
 * @index -- current index
 * @n_bd -- total number of entries in queue (must be power of 2)
 */
static
int
iwl_queue_inc_wrap(
    int index,
    int n_bd
    )
{
    return ++index & (n_bd - 1);
}

/**
 * iwl_queue_dec_wrap - decrement queue index, wrap back to end
 * @index -- current index
 * @n_bd -- total number of entries in queue (must be power of 2)
 */
static
int
iwl_queue_dec_wrap(
    int index,
    int n_bd
    )
{
    return --index & (n_bd - 1);
}

struct iwl_cmd_meta {
    /* only for SYNC commands, iff the reply skb is wanted */
    struct Iwl_Host_Cmd *source;
    u32 flags;
};
/*
 * Generic queue structure
 *
 * Contains common data for Rx and Tx queues.
 *
 * Note the difference between n_bd and n_window: the hardware
 * always assumes 256 descriptors, so n_bd is always 256 (unless
 * there might be HW changes in the future). For the normal TX
 * queues, n_window, which is the size of the software queue data
 * is also 256; however, for the command queue, n_window is only
 * 32 since we don't need so many commands pending. Since the HW
 * still uses 256 BDs for DMA though, n_bd stays 256. As a result,
 * the software buffers (in the variables @meta, @txb in struct
 * iwl_txq) only have 32 entries, while the HW buffers (@tfds in
 * the same struct) have 256.
 * This means that we end up with the following:
 *  HW entries: | 0 | ... | N * 32 | ... | N * 32 + 31 | ... | 255 |
 *  SW entries:           | 0      | ... | 31          |
 * where N is a number between 0 and 7. This means that the SW
 * data is a window overlayed over the HW queue.
 */
struct iwl_queue {
    int n_bd;              /* number of BDs in this queue */
    int write_ptr;       /* 1-st empty entry (index) host_w*/
    int read_ptr;         /* last used entry (index) host_r*/
    /* use for monitoring and recovering the stuck queue */
    dma_addr_t dma_addr;   /* physical addr for BD's */
    int n_window;          /* safe queue window */
    u32 id;
    int low_mark;          /* low watermark, resume queue if free
                * space more than this */
    int high_mark;         /* high watermark, stop queue if free
                * space less than this */
};

#define TFD_TX_CMD_SLOTS 256
#define TFD_CMD_SLOTS 32

/*
 * The FH will write back to the first TB only, so we need
 * to copy some data into the buffer regardless of whether
 * it should be mapped or not. This indicates how big the
 * first TB must be to include the scratch buffer. Since
 * the scratch is 4 bytes at offset 12, it's 16 now. If we
 * make it bigger then allocations will be bigger and copy
 * slower, so that's probably not useful.
 */
#define IWL_HCMD_SCRATCHBUF_SIZE    16

struct iwl_pcie_txq_entry {
    struct iwl_device_cmd *cmd;
    SK_BUFF *skb;
    struct iwl_tx_skb tx_skb;
    /* buffer to free after command completes */
    const void *free_buf;
    struct iwl_cmd_meta meta;
    UINT64* tb_cpuaddr[IWL_NUM_OF_TBS];
    dma_addr_t tb_phys[IWL_NUM_OF_TBS];
};

/**
 * struct iwl_txq - Tx Queue for DMA
 * @q: generic Rx/Tx queue descriptor
 * @tfds: transmit frame descriptors (DMA memory)
 * @scratchbufs: start of command headers, including scratch buffers, for
 *	the writeback -- this is DMA memory and an array holding one buffer
 *	for each command on the queue
 * @scratchbufs_dma: DMA address for the scratchbufs start
 * @entries: transmit entries (driver state)
 * @lock: queue lock
 * @stuck_timer: timer that fires if queue gets stuck
 * @trans_pcie: pointer back to transport (for timer)
 * @need_update: indicates need to update read/write index
 * @active: stores if queue is active
 * @ampdu: true if this queue is an ampdu queue for an specific RA/TID
 *
 * A Tx queue consists of circular buffer of BDs (a.k.a. TFDs, transmit frame
 * descriptors) and required locking structures.
 */
struct iwl_txq {
    struct iwl_queue q;
    struct iwl_tfd *tfds;
    void *tfds_mapping;
    void *tbs_mapping[TFD_QUEUE_SIZE_MAX][IWL_NUM_OF_TBS];
    struct iwl_pcie_txq_scratch_buf *scratchbufs;
    void *scratch_buf_mapping;
    u64 scratchbufs_dma;
    struct iwl_pcie_txq_entry *entries;

    struct iwl_trans_pcie *trans_pcie;
    u8 need_update;
    u8 active;
    bool ampdu;
};




/**
 * struct iwl_trans_pcie - PCIe transport specific data
 * @rxq: all the RX queue data
 * @rx_replenish: work that will be called when buffers need to be allocated
 * @drv - pointer to iwl_drv
 * @trans: pointer to the generic transport area
 * @scd_base_addr: scheduler sram base address in SRAM
 * @scd_bc_tbls: pointer to the byte count table of the scheduler
 * @kw: keep warm address
 * @pci_dev: basic pci-network driver stuff
 * @hw_base: pci hardware address support
 * @ucode_write_complete: indicates that the ucode has been copied.
 * @ucode_write_waitq: wait queue for uCode load
 * @cmd_queue - command queue number
 * @rx_buf_size_8k: 8 kB RX buffer size
 * @bc_table_dword: true if the BC table expects DWORD (as opposed to bytes)
 * @rx_page_order: page order for receive buffer size
 * @wd_timeout: queue watchdog timeout (jiffies)
 * @reg_lock: protect hw register access
 * @cmd_in_flight: true when we have a host command in flight
 */

typedef struct iwl_trans_pcie {
    struct iwl_rxq rxq;
    IWL_TRANS *Trans;
    INT32 Ict_Index;
    BOOLEAN Use_Ict;
    struct isr_statistics isr_stats;
    UINT32 Inta_Mask;
    UINT32 Scd_Base_Addr;
    struct iwl_dma_ptr scd_bc_tbls;
    struct iwl_dma_ptr kw;

    struct iwl_txq *txq;
    unsigned long queue_used[BITS_TO_LONGS(IWL_MAX_HW_QUEUES)];
    unsigned long queue_stopped[BITS_TO_LONGS(IWL_MAX_HW_QUEUES)];

    BOOLEAN Ucode_Write_Complete;

    UINT8 Cmd_Queue;
    UINT8 cmd_fifo;
    UINT8 n_no_reclaim_cmds;
    UINT8 no_reclaim_cmds[MAX_NO_RECLAIM_CMDS];

    BOOLEAN rx_buf_size_8k;
    BOOLEAN bc_table_dword;
    UINT32 rx_page_order;

    const char *const *command_names;

    /* queue watchdog */
    unsigned long wd_timeout;
    BOOLEAN cmd_in_flight;
}IWL_TRANS_PCIE;


#define IWL_TRANS_GET_PCIE_TRANS(IwlTransContext) \
    ((struct iwl_trans_pcie *) (IwlTransContext->TransPcie))

//########################
//Read Write and DMA calls

static
bool
iwl_queue_used(
    const struct iwl_queue *q,
    int i
    )
{
    return q->write_ptr >= q->read_ptr ?
           (i >= q->read_ptr && i < q->write_ptr) :
           !(i < q->read_ptr && i >= q->write_ptr);
}

static
const
char *get_cmd_string(
    struct iwl_trans_pcie *trans_pcie,
    u8 cmd
    )
{
    if (!trans_pcie->command_names || !trans_pcie->command_names[cmd])
        return "UNKNOWN";
    return trans_pcie->command_names[cmd];
}


static
u8
get_cmd_index(
    struct iwl_queue *q,
    u32 index
    )
{
    return (u8)(index & (q->n_window - 1));
}



void
Iwl_Trans_Pcie_Free(
    struct iwl_trans_context *IwlTransContext
    );

/*****************************************************
* RX
******************************************************/

INT32
Iwl_Pcie_Rx_Init(
    IN IWL_TRANS_CONTEXT *IwlTransContext
    );
INT32
iwl_pcie_rx_stop(
    struct iwl_trans_context *IwlTransContext
    );

void
iwl_pcie_rx_free(
    struct iwl_trans_context *IwlTransContext
    );

/*****************************************************
* TX / HCMD
******************************************************/
int
iwl_pcie_tx_init(
    struct iwl_trans_context *IwlTransContext
    );

void
iwl_pcie_tx_start(
    struct iwl_trans_context *IwlTransContext,
    u32 scd_base_addr
    );

int
iwl_pcie_tx_stop(
    struct iwl_trans_context *IwlTransContext
    );

VOID
Iwl_Trans_Pcie_Txq_Enable (
    IN struct iwl_trans_context *IwlTransContext,
    IN UINT32 txq_id,
    IN UINT32 fifo,
    IN INTN sta_id,
    IN UINT32 tid,
    IN UINT32 frame_limit,
    IN UINT16 ssn
    );
void
iwl_trans_pcie_txq_disable(
    struct iwl_trans_context *IwlTransContext,
    int txq_id
    );

void
iwl_trans_pcie_reclaim(
    struct iwl_trans_context *IwlTransContext,
    int txq_id,
    int ssn,
    struct iwl_tx_skb *skb
    );

void
iwl_pcie_tx_free(
    struct iwl_trans_context *IwlTransContext
    );

void
Iwl_Trans_Pcie_Configure(
    struct iwl_trans_context *IwlTransContext,
    const struct iwl_trans_config *trans_cfg
    );

/*****************************************************
* Error handling
******************************************************/

/*****************************************************
* Helpers
******************************************************/

INTN
Iwl_Trans_Pcie_Write_Mem(
    struct iwl_trans_context *IwlTransContext,
    u32 addr,
    const void *buf,
    INTN dwords
    );

INTN
Iwl_Trans_Pcie_Read_Mem(
    struct iwl_trans_context *IwlTransContext,
    u32 addr,
    void *buf,
    INTN dwords
    );

INT32
Dma_Map (
    IN IWL_TRANS_CONTEXT *IwlTransContext,
    IN VOID *CpuAddr,
    IN UINT32 NumBytes,
    IN UINT32 Direction,
    IN OUT UINT64 *DeviceAddrPtr,
    OUT VOID      **mapping
    );

INTN
iwl_trans_pcie_send_hcmd(
    struct iwl_trans_context *IwlTransContext,
    struct Iwl_Host_Cmd *cmd
    );

void
iwl_stop_queue(
    struct iwl_trans_context *IwlTransContext,
    struct iwl_txq *txq
    );

INT32
Dma_Unmap (
    IN IWL_TRANS_CONTEXT *IwlTransContext,
    IN VOID *mapping
    );

INT32
Iwl_Trans_Pcie_Start_Hw(
    IN IWL_TRANS_CONTEXT *IwlTransContext
    );

INT32
Iwl_Trans_Pcie_Start_Fw(
    IN IWL_TRANS_CONTEXT *IwlTransContext,
    const struct fw_img *fw,
    BOOLEAN run_in_rfkill
    );

VOID
Iwl_Trans_Pcie_Write32(
    IN IWL_TRANS_CONTEXT *IwlTransContext,
    UINT32 Ofs,
    UINT32 Val
    );

UINT32
Iwl_Trans_Pcie_Read32(
    IN IWL_TRANS_CONTEXT *IwlTransContext,
    UINT32 Ofs
    );

VOID
Iwl_Trans_Pcie_Set_Bits_Mask(
    IN IWL_TRANS_CONTEXT *IwlTransContext,
    UINT32 Reg,
    UINT32 Mask,
    UINT32 Value
    );

VOID
Iwl_Trans_Pcie_Fw_Alive(
    IWL_TRANS_CONTEXT *IwlTransContext,
    u32 scd_addr
    );

VOID
Iwl_Trans_Pcie_Stop_Device(
    IN IWL_TRANS_CONTEXT *IwlTransContext
    );

INT32
test_bit (
    IN UINT64 nr,
    IN const volatile void *addr
    );

INT32
test_and_set_bit (
    IN UINT64 nr,
    IN unsigned long *addr
    );
INT32
test_and_clear_bit (
    IN UINT64 nr,
    IN unsigned long  *addr
    );
VOID
clear_bit (
    INTN nr,
    unsigned long *addr
    );

VOID
set_bit (
    UINT64 nr,
    unsigned long *addr
    );


VOID
iwl_write_direct32(
    struct iwl_trans_context *IwlTransContext,
    u32 reg,
    u32 value
    );

UINT32
IwlReadDirect32 (
    IN IWL_TRANS_CONTEXT *IwlTransContext,
    IN u32 reg
    );

VOID
Iwl_Trans_Write8(
    IN IWL_TRANS_CONTEXT *IwlTransContext,
    UINT32 ofs,
    UINT8 Val
    );

UINT32
IwlTransPcieReadPrph (
    IN IWL_TRANS_CONTEXT *IwlTransContext,
    IN UINT32 Ofs
    );

VOID
IwlTransPcieWritePrph (
    IN IWL_TRANS_CONTEXT *IwlTransContext,
    IN UINT32 Ofs,
    IN UINT32 Val);

BOOLEAN
IwlTransPcieGrabNicAccess (
    IN IWL_TRANS_CONTEXT *IwlTransContext,
    IN BOOLEAN Silent
    );

VOID
IwlTransPcieReleaseNicAccess (
    IN IWL_TRANS_CONTEXT *IwlTransContext
    );


INTN
iwl_trans_pcie_tx(
    struct iwl_trans_context *IwlTransContext,
    SK_BUFF *skb,
    struct iwl_device_cmd *dev_cmd,
    INTN txq_id
    );


STATIC
VOID
Iwl_Pcie_Rxq_Free_Rbs(
    IN IWL_TRANS_CONTEXT *IwlTransContext
    );

STATIC
VOID
Iwl_Pcie_Rx_Init_Rxb_Lists(
    struct iwl_rxq *rxq
    );

STATIC
VOID
Iwl_Pcie_Rx_Replenish(
    IN IWL_TRANS_CONTEXT *IwlTransContext
    );

STATIC
VOID
Iwl_Pcie_Rx_Hw_Init(
    IN IWL_TRANS_CONTEXT *IwlTransContext,
    struct iwl_rxq *rxq
    );

STATIC
VOID
Iwl_Pcie_Rxq_Inc_Wr_Ptr(
    IN IWL_TRANS_CONTEXT *IwlTransContext,
    struct iwl_rxq *rxq
    );

VOID
Iwl_Trans_Pcie_Rf_Kill(
    IN IWL_TRANS_CONTEXT *IwlTransContext,
    BOOLEAN state
    );

VOID
Iwl_Enable_Rfkill_Int(
    IN IWL_TRANS_CONTEXT *IwlTransContext
    );

BOOLEAN
Iwl_Is_Rfkill_Set(
    IN IWL_TRANS_CONTEXT *IwlTransContext
    );

VOID
Iwl_Disable_Interrupts (
    IN IWL_TRANS_CONTEXT *IwlTransContext
    );

VOID
Iwl_Enable_Interrupts(
    IN IWL_TRANS_CONTEXT *IwlTransContext
    );

VOID
Iwl_Trans_Pcie_Rf_Kill(
    IN IWL_TRANS_CONTEXT *IwlTransContext,
    BOOLEAN state
    );

VOID
iwl_pcie_hcmd_complete(
    IN struct iwl_trans_context *IwlTransContext,
    IN struct iwl_rx_cmd_buffer *rxb,
    IN int handler_status
    );

VOID
Iwl_Pcie_Rx_Handle(
    IN IWL_TRANS_CONTEXT *IwlTransContext
    );

VOID
Iwl_Pcie_Irq_Handler(
    IN IWL_TRANS_CONTEXT *IwlTransContext
    );

EFI_STATUS
Iwl_Trans_Pcie_Initialize(IWL_TRANS_CONTEXT *iwlTransPcieCtxt);

STATIC
INTN
iwl_trans_read_mem (
    IN struct iwl_trans_context *IwlTransContext,
    IN UINT32 addr,
    VOID *buf,
    INTN dwords
    )
{
    return Iwl_Trans_Pcie_Read_Mem (IwlTransContext,addr,buf,dwords);
}

STATIC
UINT32
iwl_trans_read_mem32 (
    IN struct iwl_trans_context *IwlTransContext,
    IN UINT32 addr
    )
{
    UINT32 value;

    if ((iwl_trans_read_mem (IwlTransContext,addr,&value,1)))
        return 0xa5a5a5a5;

    return value;
}

STATIC
INTN
iwl_trans_write_mem (
    IN struct iwl_trans_context *IwlTransContext,
    IN UINT32 addr,
    IN const VOID *buf,
    IN INTN dwords
    )
{
    return Iwl_Trans_Pcie_Write_Mem(IwlTransContext,addr,buf,dwords);
}

STATIC
INTN
iwl_trans_write_mem32 (
    IN struct iwl_trans_context *IwlTransContext,
    UINT32 addr,
    UINT32 val
    )
{
    return iwl_trans_write_mem (IwlTransContext,addr,&val,1);
}

#endif
