/**
 * @file rx.c
 *
 * @details
 * This header file contains the code to implement receive
 * operation for Intel UNDI driver.
 *
 * @copyright
 * Copyright (c) 2017 Dell Inc. All rights reserved. This
 * software and associated documentation (if any) is furnished
 * under a license and may only be used or copied in accordance
 * with the terms of the license. Except as permitted by such
 * license, no part of this software or documentation may be
 * reproduced, stored in a retrieval system, or transmitted in
 * any form or by any means without the express written consent
 * of Dell Inc.
 */

/******************************************************************************
 *
 * Copyright(c) 2003 - 2014 Intel Corporation. All rights reserved.
 *
 * Portions of this file are derived from the ipw3945 project, as well
 * as portions of the ieee80211 subsystem header files.
 *
 * This program is free software; you can redistribute it and/or modify it
 * under the terms of version 2 of the GNU General Public License as
 * published by the Free Software Foundation.
 *
 * This program is distributed in the hope that it will be useful, but WITHOUT
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
 * FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License for
 * more details.
 *
 * You should have received a copy of the GNU General Public License along with
 * this program; if not, write to the Free Software Foundation, Inc.,
 * 51 Franklin Street, Fifth Floor, Boston, MA 02110, USA
 *
 * The full GNU General Public License is included in this distribution in the
 * file called LICENSE.
 *
 * Contact Information:
 *  Intel Linux Wireless <ilw@linux.intel.com>
 * Intel Corporation, 5200 N.E. Elam Young Parkway, Hillsboro, OR 97124-6497
 *
 *****************************************************************************/


#include "internal.h"
#include "WirelessMemoryManager.h"
#include "IwlMvmOps.h"
#include <Uefi.h>
#include <Library/DebugLib.h>


#define list_entry(ptr, type, member) \
    BASE_CR(ptr, type, member)

#define list_first_entry(ptr, type, member) \
    list_entry((ptr)->next, type, member)

#define DMA_BIT_MASK(n) (((n) == 64) ? ~0ULL : ((1ULL<<(n))-1))
int
list_empty(
    const struct list_head *head
    )
{
    return head->next == head;
}

void
INIT_LIST_HEAD(
    struct list_head *list
    )
{
    if(list != NULL)
    {
        list->next = list;
        list->prev = list;
    }
}

/*
 * Insert a new entry between two known consecutive entries.
 *
 * This is only for internal list manipulation where we know
 * the prev/next entries already!
 */
void
__list_add(
    struct list_head *new,
    struct list_head *prev,
    struct list_head *next
    )
{
    if((new != NULL) && (prev != NULL) && (next != NULL))
    {
        next->prev = new;
        new->next = next;
        new->prev = prev;
        prev->next = new;
    }
}

/**
 * list_add - add a new entry
 * @new: new entry to be added
 * @head: list head to add it after
 *
 * Insert a new entry after the specified head.
 * This is good for implementing stacks.
 */
void
list_add(
    struct list_head *new,
    struct list_head *head
    )
{
    __list_add(new, head, head->next);
}


/**
 * list_add_tail - add a new entry
 * @new: new entry to be added
 * @head: list head to add it before
 *
 * Insert a new entry before the specified head.
 * This is useful for implementing queues.
 */
void
list_add_tail(
    struct list_head *new,
    struct list_head *head
    )
{
    __list_add(new, head->prev, head);
}

void
__list_del(
    struct list_head * prev,
    struct list_head * next
    )
{
    if((prev != NULL) && (next != NULL))
    {
        next->prev = prev;
        prev->next = next;
    }
}


void
list_del(
    struct list_head *entry
    )
{
    __list_del(entry->prev, entry->next);
}


/*******************************************************************/
/*                                      RX Calls implemented here.                     */
/*******************************************************************/

/*
 * iwl_rxq_space - Return number of free slots available in queue.
 */
INT32
iwl_rxq_space(
    const struct iwl_rxq *rxq
    )
{
    /* Make sure RX_QUEUE_SIZE is a power of 2 */

    /*
     * There can be up to (RX_QUEUE_SIZE - 1) free slots, to avoid ambiguity
     * between empty and completely full queues.
     * The following is equivalent to modulo by RX_QUEUE_SIZE and is well
     * defined for negative dividends.
     */
    return (rxq->read - rxq->write - 1) & (RX_QUEUE_SIZE - 1);
}

/*
 * iwl_dma_addr2rbd_ptr - convert a DMA address to a uCode read buffer ptr
 */
UINT32
iwl_pcie_dma_addr2rbd_ptr(
    dma_addr_t dma_addr
    )
{
    return ((u32)(dma_addr >> 8));
}



/*
 * iwl_pcie_rx_stop - stops the Rx DMA
 */
INT32
iwl_pcie_rx_stop(
    struct iwl_trans_context *IwlTransContext
    )
{
    iwl_write_direct32(IwlTransContext, FH_MEM_RCSR_CHNL0_CONFIG_REG, 0);
    return (IwlPollDirectBit(IwlTransContext, FH_MEM_RSSR_RX_STATUS_REG,
                             FH_RSSR_CHNL0_RX_STATUS_CHNL_IDLE, 1000));
}

VOID
Iwl_Trans_Fw_Error(
        IN IWL_TRANS_CONTEXT *IwlTransContext
    )
{
    if(IwlTransContext != NULL)
    {
        /* prevent double restarts due to the same erroneous FW */
        if (!test_and_set_bit(STATUS_FW_ERROR, &IwlTransContext->Status))
        {
            Iwl_Mvm_Nic_Error_Stub();
        }
    }
}

VOID
Iwl_Pcie_Irq_Handle_Error(
        IN IWL_TRANS_CONTEXT *IwlTransContext
    )
{
    if(IwlTransContext != NULL)
    {
        /* The STATUS_FW_ERROR bit is set in this function. This must happen
         * before we wake up the command caller, to ensure a proper cleanup. */
        Iwl_Trans_Fw_Error(IwlTransContext);
        clear_bit(STATUS_SYNC_HCMD_ACTIVE, &IwlTransContext->Status);
    }
}

VOID
Iwl_Pcie_Rxq_Inc_Wr_Ptr(
    struct iwl_trans_context *IwlTransContext,
    struct iwl_rxq *rxq
    )
{
    UINT32 reg;

    if(!IwlTransContext || !rxq)
    {
        goto out;
    }
    if (rxq->need_update == 0)
        goto out;

    /*
     * explicitly wake up the NIC if:
     * 1. shadow registers aren't enabled
     * 2. there is a chance that the NIC is asleep
     */
    if (!IwlTransContext->IwlConfig->Base_Params->Shadow_Reg_Enable &&
        test_bit(STATUS_TPOWER_PMI, &IwlTransContext->Status))
    {
        reg = Iwl_Trans_Pcie_Read32(IwlTransContext, CSR_UCODE_DRV_GP1);

        if (reg & CSR_UCODE_DRV_GP1_BIT_MAC_SLEEP) {
            DEBUG( (EFI_D_ERROR, "Rx queue requesting wakeup, GP1 = 0x%x\n", reg ));
            Iwl_Trans_Pcie_Set_Bits_Mask(IwlTransContext,
                                      CSR_GP_CNTRL,
                                      CSR_GP_CNTRL_REG_FLAG_MAC_ACCESS_REQ,
                                      CSR_GP_CNTRL_REG_FLAG_MAC_ACCESS_REQ);
            goto out;
        }
    }

    rxq->write_actual = (rxq->write / 8) * 8;

    Iwl_Trans_Pcie_Write32(IwlTransContext, FH_RSCSR_CHNL0_WPTR, rxq->write_actual);
    rxq->need_update = 0;

out:
    return;

}


VOID
Iwl_Pcie_Rxq_Alloc_Rbs(
        IWL_TRANS_CONTEXT *IwlTransContext
    )
{
    EFI_STATUS Status;
    INT32 retStatus = 0;
    struct iwl_trans_pcie *trans_pcie = NULL;
    struct iwl_rxq *rxq = NULL;
    struct iwl_rx_mem_buffer *rxb = NULL;

    if(!IwlTransContext)
    {
        return;
    }

    trans_pcie = IWL_TRANS_GET_PCIE_TRANS(IwlTransContext);
    rxq = &trans_pcie->rxq;

    while(1)
    {
        if(list_empty(&rxq->rx_used))
            return;
        rxb = list_first_entry(&rxq->rx_used, struct iwl_rx_mem_buffer, list);
        list_del(&rxb->list);

        /*Requesting a page from the pre-allocated region*/
        Status =  WirelessAllocatePage((VOID **)&rxb->cpu_addr);
        if (EFI_ERROR(Status))
        {
            DEBUG( (EFI_D_ERROR, "Iwl_Pcie_Rxq_Alloc_Rbs:No memory can be allocated for cpu_addr,status = %d\n",Status ));
        }
        if(!rxb->cpu_addr) {
            if((rxq->free_count <= RX_LOW_WATERMARK))
                DEBUG( (EFI_D_ERROR, "Free buffers reached less than low watermark.\n"));
            return;
        }

        retStatus = Dma_Map(IwlTransContext,
                            rxb->cpu_addr,
                            PAGE_SIZE,
                            FROM_DEVICE,
                            &rxb->page_dma,
                            &rxb->Mapping);

        if(retStatus != 0)
        {
            DEBUG( (EFI_D_ERROR, "Alloc Rbs: Dma Map failed,return status = %d\n",retStatus ));

            /* Returning the page to the memory manager*/
            Status =  WirelessFreePage(rxb->cpu_addr);
            if (EFI_ERROR(Status)) {
                DEBUG( (EFI_D_ERROR, "Iwl_Pcie_Rxq_Alloc_Rbs: Memory release failed,status = %d\n",Status ));
            }
            rxb->cpu_addr = NULL;
        }
#if 0
        if(rxb->page_dma & ~DMA_BIT_MASK(36))
            DEBUG( EFI_D_ERROR, (" In Rx alloc RX no properly aligned" ));
        /* and also 256 byte aligned! */
        if(rxb->page_dma & DMA_BIT_MASK(8))
            DEBUG( EFI_D_ERROR, ("In Rx alloc Rx not 256 byte aligned\n " ));
#endif

        list_add_tail(&rxb->list, &rxq->rx_free);
        rxq->free_count++;
    }
}

VOID
Iwl_Pcie_Rxq_Restock(
     IWL_TRANS_CONTEXT  *IwlTransContext
    )
{
    struct iwl_trans_pcie *trans_pcie = NULL;
    struct iwl_rxq *rxq = NULL;
    struct iwl_rx_mem_buffer *rxb;
    UINT32 *temp;
    INT32 retStatus = 0;

    if(!IwlTransContext)
    {
        return;
    }

    trans_pcie = IWL_TRANS_GET_PCIE_TRANS(IwlTransContext);
    rxq = &trans_pcie->rxq;
    /*
     * If the device isn't enabled - not need to try to add buffers...
     * This can happen when we stop the device and still have an interrupt
     * pending. We stop the APM before we sync the interrupts because we
     * have to (see comment there). On the other hand, since the APM is
     * stopped, we cannot access the HW (in particular not prph).
     * So don't try to restock if the APM has been already stopped.
     */
    if (!test_bit(STATUS_DEVICE_ENABLED, &IwlTransContext->Status))
        return;

    retStatus = Dma_Unmap(IwlTransContext, rxq->Mapping_bd);
    if (retStatus < 0)
    {
        DEBUG( (EFI_D_ERROR, "Iwl_Pcie_Rxq_Restock: Dma Unmap Failed,status = %d\n",retStatus ));
    }

    while ((iwl_rxq_space(rxq) > 0) && (rxq->free_count)) {
        /* The overwritten rxb must be a used one */
        rxb = rxq->queue[rxq->write];
        if((rxb && rxb->cpu_addr))
            DEBUG( (EFI_D_ERROR, "unused frame in restock\n" ));
        /* Get next free Rx buffer, remove from free list */
        rxb = list_first_entry(&rxq->rx_free, struct iwl_rx_mem_buffer, list);
        list_del(&rxb->list);

        /* Point to Rx buffer via next RBD in circular buffer */
        temp = (UINT32 *)rxq->bd;
        temp[rxq->write] = iwl_pcie_dma_addr2rbd_ptr(rxb->page_dma);
        rxq->queue[rxq->write] = rxb;
        rxq->write = (rxq->write + 1) & RX_QUEUE_MASK;
        rxq->free_count--;
    }

    /* If we've added more space for the firmware to place data, tell it.
     * Increment device's write pointer in multiples of 8. */
    if (rxq->write_actual != (rxq->write & ~0x7)) {
        rxq->need_update = 1;
        Iwl_Pcie_Rxq_Inc_Wr_Ptr(IwlTransContext, rxq);
    }

    retStatus = Dma_Map(IwlTransContext,
                                 rxq->bd,
                                 PAGE_SIZE,
                                 TO_DEVICE,
                                 &rxq->bd_dma,
                                 &rxq->Mapping_bd);
    if (retStatus != 0) {
        DEBUG( (EFI_D_ERROR, "Iwl_Pcie_Rxq_Restock : unable to do DMA MAP, status = %d",retStatus ));
        return;
    }
}


VOID
Iwl_Pcie_Rx_Replenish(
    IWL_TRANS_CONTEXT  *IwlTransContext
    )
{

    Iwl_Pcie_Rxq_Alloc_Rbs(IwlTransContext);

    Iwl_Pcie_Rxq_Restock(IwlTransContext);

}

static VOID
Iwl_Pcie_Get_Rxcb(
    struct iwl_trans_context *IwlTransContext,
    struct iwl_rx_cmd_buffer *rxcb,
    UINT64 *cpu_addr,
    UINTN offset)
{
    struct iwl_trans_pcie *trans_pcie = NULL;
    UINTN max_len = 0;

    if((IwlTransContext != NULL) && (rxcb != NULL))
    {
        trans_pcie = IWL_TRANS_GET_PCIE_TRANS(IwlTransContext);
        max_len = (UINTN)(((UINT64)PAGE_SIZE) << trans_pcie->rx_page_order);
        rxcb->cpu_addr = cpu_addr;
        rxcb->_offset = offset;
        rxcb->_page_stolen = false;
        rxcb->_rx_page_order = trans_pcie->rx_page_order;
        rxcb->truesize = max_len;
    }
}


static EFI_STATUS
Iwl_Pcie_Get_And_Validate_RxPacket(
    struct iwl_rx_mem_buffer *rxb,
    UINTN offset,
    struct Iwl_Rx_Packet **pkt,
    INT32 *len)

{
    if(!rxb || !pkt)
    {
        return EFI_INVALID_PARAMETER;
    }
    *pkt = (struct Iwl_Rx_Packet *)((UINT8 *)rxb->cpu_addr + offset);
    if ((*pkt)->len_n_flags == CONV_TO_LE32_BIT(FH_RSCSR_FRAME_INVALID))
    {
        return EFI_NO_RESPONSE;
    }
    *len = le32_to_cpu((*pkt)->len_n_flags) & FH_RSCSR_FRAME_SIZE_MSK;
    *len += sizeof(UINT32);/* account for status word */
    return EFI_SUCCESS;
}

static BOOLEAN
CheckForCommnadQueueReclaim(
        struct iwl_trans_context *IwlTransContext,
        struct Iwl_Rx_Packet *pkt)
{

    BOOLEAN reclaim;
    struct iwl_trans_pcie *trans_pcie = IWL_TRANS_GET_PCIE_TRANS(IwlTransContext);

    /* Reclaim a command buffer only if this packet is a response
     *   to a (driver-originated) command.
     * If the packet (e.g. Rx frame) originated from uCode,
     *   there is no command buffer to reclaim.
     * Ucode should set SEQ_RX_FRAME bit if ucode-originated,
     *   but apparently a few don't get set; catch them here. */
    reclaim = !(pkt->Hdr.Sequence & SEQ_RX_FRAME);
    if(reclaim) {
        int i;

        for (i = 0; i < trans_pcie->n_no_reclaim_cmds; i++) {
            if (trans_pcie->no_reclaim_cmds[i] == pkt->Hdr.Cmd)
            {
                reclaim = false;
                break;
            }
        }
    }
    return reclaim;
}

static VOID
Iwl_Pcie_ProcessResponse_ForCmdQueue_Cmds(
    struct iwl_trans_context *IwlTransContext,
    BOOLEAN reclaim,
    INT32 cmd_index,
    struct iwl_rx_cmd_buffer *rxb,
    INT32 handler_status
    )
{
    struct iwl_trans_pcie *trans_pcie = NULL;
    struct iwl_txq *txq = NULL;

    if(!IwlTransContext || !rxb)
    {
        return;
    }
    trans_pcie = IWL_TRANS_GET_PCIE_TRANS(IwlTransContext);
    txq = &trans_pcie->txq[trans_pcie->Cmd_Queue];
    if (reclaim) {
        txq->entries[cmd_index].free_buf = NULL;
        /* Invoke any callbacks, transfer the buffer to caller, and fire off the (possibly) blocking
         * iwl_trans_send_cmd() as we reclaim the driver command queue
         */
        if (!rxb->_page_stolen) {
            iwl_pcie_hcmd_complete(IwlTransContext, rxb, handler_status);
        }
        else
            DEBUG( (EFI_D_ERROR, "Claim null rxb?\n" ));
    }
}
static
void
iwl_pcie_rx_handle_rb(
    struct iwl_trans_context *IwlTransContext,
    struct iwl_rx_mem_buffer *rxb
    )
{
    struct iwl_trans_pcie *trans_pcie = NULL;
    struct iwl_rxq *rxq = NULL;
    struct iwl_txq *txq = NULL;
    BOOLEAN page_stolen = false;
    UINTN max_len = 0;
    UINTN offset = 0;
    EFI_STATUS Status;
    INT32 retStatus = 0;

    if(!IwlTransContext || !rxb)
    {
        return;
    }

    trans_pcie = IWL_TRANS_GET_PCIE_TRANS(IwlTransContext);
    rxq = &trans_pcie->rxq;
    txq = &trans_pcie->txq[trans_pcie->Cmd_Queue];
    max_len = (((UINT64)PAGE_SIZE) << trans_pcie->rx_page_order);

    retStatus = Dma_Unmap(IwlTransContext, rxb->Mapping);
    if (retStatus < 0)
    {
        DEBUG( (EFI_D_ERROR, "iwl_pcie_rx_handle_rb: Dma Unmap Failed,status = %d\n",retStatus ));
    }

    while (offset + sizeof(UINT32) + sizeof(struct Iwl_Cmd_Header) < max_len) {

        struct Iwl_Rx_Packet *pkt;
        struct iwl_device_cmd *cmd;
        struct iwl_rx_cmd_buffer rxcb;
        UINT16 sequence;
        BOOLEAN reclaim;
        INT32 index, cmd_index= 0, err = 0, len;


        /* Initialize the rx command buffer with the buffer received from nic, later for dispatching it to the
         * corresponding handler.
         */
        Iwl_Pcie_Get_Rxcb(IwlTransContext,&rxcb,rxb->cpu_addr,offset);

        if(Iwl_Pcie_Get_And_Validate_RxPacket(rxb,offset,&pkt,&len) != EFI_SUCCESS)
        {
            break;
        }

        //checking if the packet received is response to Driver originated CMD.
        if(CheckForCommnadQueueReclaim(IwlTransContext,pkt) == true)
        {
            reclaim = true;

            //yes.. it is response to driver originated CMD. get the index of corresponding CMD from CMD Q.
            sequence = le16_to_cpu(pkt->Hdr.Sequence);
            index = SEQ_TO_INDEX(sequence);
            cmd_index = index & (txq->q.n_window - 1);
            cmd = txq->entries[cmd_index].cmd;
        }
        else
        {
            reclaim = false;
            cmd     = NULL;
        }

        err = Iwl_Mvm_Rx_Dispatch_Stub(&rxcb,cmd,page_stolen);

        /*Process the reclaimed command buffers*/
        Iwl_Pcie_ProcessResponse_ForCmdQueue_Cmds(IwlTransContext,reclaim,cmd_index,&rxcb,err);

        page_stolen |= rxcb._page_stolen;
        offset += ALIGN(len, FH_RSCSR_FRAME_ALIGN);
    }

    /* Except data type packets, free the receive buffer here
     * Free the command responses ,statistics and other notification packets like phy notifications and also non enqueued packets
     */
    if (!page_stolen)
    {
        if(rxb->cpu_addr != NULL) {
            Status =  WirelessFreePage(rxb->cpu_addr);
            if(EFI_ERROR(Status))
                DEBUG( (EFI_D_ERROR, "handle rb: Freepage Fail\n" ));
        }
    }

    rxb->cpu_addr = NULL;

    /*Allocate New page and queue it to free list*/
    Status =  WirelessAllocatePage((VOID **)&rxb->cpu_addr);
    if (EFI_ERROR(Status)) {
        DEBUG( (EFI_D_ERROR, "handle_rb:No memory can be allocated for cpu_addr,status = %d\n",Status ));
    }

    retStatus = Dma_Map(IwlTransContext,
                                 rxb->cpu_addr,
                                 PAGE_SIZE,
                                 FROM_DEVICE,
                                 &rxb->page_dma,
                                 &rxb->Mapping);
    if(retStatus != 0)
    {
        DEBUG( (EFI_D_ERROR, "Page Stolen  : Dma Map failed,return status = %d\n",retStatus ));
    }


#if 0
    if(rxb->page_dma & ~DMA_BIT_MASK(36))
        DEBUG( EFI_D_ERROR, (" In Replenish_Enqueue RX no properly aligned" ));
    /* and also 256 byte aligned! */
    if(rxb->page_dma & DMA_BIT_MASK(8))
        DEBUG( EFI_D_ERROR, ("In Replenish_Enqueue Rx not 256 byte aligned\n " ));
#endif

    list_add_tail(&rxb->list, &rxq->rx_free);
    rxq->free_count++;
}

/*
 * iwl_pcie_rx_handle - Main entry function for receiving responses from fw
 */
VOID
Iwl_Pcie_Rx_Handle(
        IWL_TRANS_CONTEXT *IwlTransContext
    )
{
    UINT32 r,i;
    INT32 retStatus = 0;

    struct iwl_trans_pcie *trans_pcie = NULL;
    struct iwl_rxq *rxq = NULL;

    if(!IwlTransContext)
    {
        return;
    }
    trans_pcie = IWL_TRANS_GET_PCIE_TRANS(IwlTransContext);
    rxq = &trans_pcie->rxq;

    /* uCode's read index (stored in shared DRAM) indicates the last Rx
     * buffer that the driver may process (last buffer filled by ucode).
     *
     */

    retStatus = Dma_Unmap(IwlTransContext, rxq->Mapping_rb);
    if (retStatus < 0)
    {
        DEBUG( (EFI_D_ERROR, "Iwl_Pcie_Rx_Handle: Dma Unmap Failed,status = %d\n",retStatus ));
    }

    r = (*(volatile UINT16 *)&(rxq->rb_stts->closed_rb_num)) & 0x0FFF;
    i = rxq->read;
    retStatus = Dma_Map(IwlTransContext,
                                 rxq->rb_stts,
                                 PAGE_SIZE,
                                 FROM_DEVICE,
                                 &rxq->rb_stts_dma,
                                 &rxq->Mapping_rb);
    if (retStatus != 0) {
        DEBUG( (EFI_D_ERROR, "Iwl_Pcie_Rx_Handle : unable to do DMA MAP for Mapping_rb,status = %d\n",retStatus ));
    }

    while (i != r) {
        struct iwl_rx_mem_buffer *rxb;

        rxb = rxq->queue[i];
        rxq->queue[i] = NULL;
        iwl_pcie_rx_handle_rb(IwlTransContext, rxb);

        i = (i + 1) & RX_QUEUE_MASK;
    }

    /* Backtrack one entry */
    rxq->read = i;
    Iwl_Pcie_Rxq_Restock(IwlTransContext);


}


/*
 * This is NIC Interrupt handler in linux based HW driver.
 * This needs to be called under a Timer based events in UEFI environment
 * for continuously sampling the INT registers and requesting the interrupts
 */
VOID
Iwl_Pcie_Irq_Handler(
        IWL_TRANS_CONTEXT *IwlTransContext
    )
{
    struct iwl_trans_pcie *trans_pcie = NULL;
    struct isr_statistics *isr_stats = NULL;
    UINT32 inta;
    UINT32 handled = 0;
    BOOLEAN hw_rfkill;

    if(!IwlTransContext)
    {
        return;
    }
    trans_pcie = IWL_TRANS_GET_PCIE_TRANS(IwlTransContext);
    isr_stats = &trans_pcie->isr_stats;
    //Disable the INT but don't clear.
    //After this no other interrupts will be serviced until re-enabled
    Iwl_Trans_Pcie_Write32(IwlTransContext,CSR_INT_MASK,0x00000000);

    //Read INT status register.
    inta = Iwl_Trans_Pcie_Read32(IwlTransContext, CSR_INT);

    //Check the type of interrupt and service accordingly
    if (inta & CSR_INT_BIT_HW_ERR) {
        //Tell the device to stop sending interrupts
        DEBUG( (EFI_D_ERROR, "HW ERROR\n" ));
        Iwl_Disable_Interrupts(IwlTransContext);

        isr_stats->hw++;
        Iwl_Pcie_Irq_Handle_Error(IwlTransContext);

        handled |= CSR_INT_BIT_HW_ERR;
    }

    //Interrupt to inform completion of the frame by scheduler.
    if (inta & CSR_INT_BIT_SCD) {
        isr_stats->sch++;
        handled |= CSR_INT_BIT_SCD;
    }

    //Check for Alive Interrupt by NIC
    if (inta & CSR_INT_BIT_ALIVE) {
        isr_stats->alive++;
        handled |= CSR_INT_BIT_ALIVE;
    }

    //Check Radio enable / disable of NIC and cleanup the resources accordingly in PCI layer.
    if (inta & CSR_INT_BIT_RF_KILL) {
        hw_rfkill = Iwl_Is_Rfkill_Set(IwlTransContext);

        DEBUG( (EFI_D_INFO, "RF_KILL bit toggled to %x.\n",hw_rfkill ));
        isr_stats->rfkill++;
        Iwl_Trans_Pcie_Rf_Kill(IwlTransContext, hw_rfkill);
        if (hw_rfkill) {
            set_bit(STATUS_RFKILL, &IwlTransContext->Status);
            if (test_and_clear_bit(STATUS_SYNC_HCMD_ACTIVE,&IwlTransContext->Status))
                DEBUG( (EFI_D_INFO, "Rfkill while SYNC HCMD in flight\n" ));
        }
        else{
            clear_bit(STATUS_RFKILL, &IwlTransContext->Status);
        }

        handled |= CSR_INT_BIT_RF_KILL;
    }

    // Chip got too hot and stopped itself
    if (inta & CSR_INT_BIT_CT_KILL) {
        DEBUG( (EFI_D_INFO, "Microcode CT kill error detected.\n" ));
        isr_stats->ctkill++;
        handled |= CSR_INT_BIT_CT_KILL;
    }

    // Error detected by uCode
    if (inta & CSR_INT_BIT_SW_ERR) {

        isr_stats->sw++;
        DEBUG( (EFI_D_ERROR, "---- Error detected by uCode-------\n" ));
        //Iwl_Pcie_Irq_Handle_Error(Trans);
        handled |= CSR_INT_BIT_SW_ERR;
    }

    //uCode wakes up after power-down sleep
    //As of now not powering down the NIC. Will update this functionality based on requirement.
    if (inta & CSR_INT_BIT_WAKEUP) {
        DEBUG( (EFI_D_INFO, "Interrupt occured for Wakeup\n" ));
        isr_stats->wakeup++;
        handled |= CSR_INT_BIT_WAKEUP;
    }

    //All uCode command responses, including Tx command responses,
    //* Rx "responses" (frame-received notification), and other
    //* notifications from uCode come through here

    if (inta & (CSR_INT_BIT_FH_RX | CSR_INT_BIT_SW_RX | CSR_INT_BIT_RX_PERIODIC)) {

        if (inta & (CSR_INT_BIT_FH_RX | CSR_INT_BIT_SW_RX)) {
            handled |= (CSR_INT_BIT_FH_RX | CSR_INT_BIT_SW_RX);
            Iwl_Trans_Pcie_Write32(IwlTransContext, CSR_FH_INT_STATUS, CSR_FH_INT_RX_MASK);
        }

        if (inta & CSR_INT_BIT_RX_PERIODIC) {
            handled |= CSR_INT_BIT_RX_PERIODIC;
            Iwl_Trans_Pcie_Write32(IwlTransContext, CSR_INT, CSR_INT_BIT_RX_PERIODIC);
        }

        //Disable periodic interrupt; we use it as just a one-shot.
        //Trans->Ops->Write8(Trans, CSR_INT_PERIODIC_REG, CSR_INT_PERIODIC_DIS);

        Iwl_Pcie_Rx_Handle(IwlTransContext);



        //* Enable periodic interrupt in 8 msec only if we received
        //* real RX interrupt (instead of just periodic int), to catch
        //* any dangling Rx interrupt.  If it was just the periodic
        //* interrupt, there was no dangling Rx activity, and no need
        //* to extend the periodic interrupt; one-shot is enough.

        if (inta & (CSR_INT_BIT_FH_RX | CSR_INT_BIT_SW_RX))
            ;//Trans->Ops->Write8(Trans, CSR_INT_PERIODIC_REG, CSR_INT_PERIODIC_ENA);

        isr_stats->rx++;
    }

    //   This "Tx" DMA channel is used only for loading uCode
    if (inta & CSR_INT_BIT_FH_TX) {
        Iwl_Trans_Pcie_Write32(IwlTransContext, CSR_FH_INT_STATUS, CSR_FH_INT_TX_MASK);
        isr_stats->tx++;
        handled |= CSR_INT_BIT_FH_TX;
        //Wake up uCode load routine, now that load is complete
        trans_pcie->Ucode_Write_Complete = true;
    }


    if (inta & ~handled) {
        isr_stats->unhandled++;
    }

    if (inta & ~(trans_pcie->Inta_Mask)) {
        //PRINT(L"Disabled INTA bits 0x%08x were pending\n",
        //       inta & ~trans_pcie->Inta_Mask);
    }

    /* Re-enable all interrupts */
    /* only Re-enable if disabled by irq */
    Iwl_Trans_Pcie_Write32(IwlTransContext, CSR_INT, handled);
    if (!test_bit(STATUS_INT_ENABLED, &IwlTransContext->Status))
        Iwl_Enable_Interrupts(IwlTransContext);
    /* Re-enable RF_KILL if it occurred */
    if (handled & CSR_INT_BIT_RF_KILL)
        Iwl_Enable_Rfkill_Int(IwlTransContext);

    return;
}

INT32
Iwl_Pcie_Rx_Alloc(
        IWL_TRANS_CONTEXT *IwlTransContext
    )
{
    struct iwl_trans_pcie *trans_pcie = NULL;
    struct iwl_rxq *rxq = NULL;
    EFI_STATUS Status;
    INT32 retStatus = 0;

    if(!IwlTransContext)
    {
        return -1;
    }
    trans_pcie = IWL_TRANS_GET_PCIE_TRANS(IwlTransContext);
    rxq = &trans_pcie->rxq;
    SetMem(&trans_pcie->rxq, sizeof(trans_pcie->rxq),0);

    //DMA Memory allocation for RX queue size
    //Get Virtual address of the allocate memory location

    Status =  WirelessAllocatePage((VOID **)&rxq->bd);

    if (EFI_ERROR(Status)) {
        DEBUG( (EFI_D_ERROR, "Iwl_Pcie_Rx_Alloc: rxq->bd Memory Allocation Failed,status = %d\n",Status ));
        goto error;
    }

    //Get virtual and Physical Address of the Page

    retStatus = Dma_Map(IwlTransContext,
                                 rxq->bd,
                                 PAGE_SIZE,
                                 TO_DEVICE,
                                 &rxq->bd_dma,
                                 &rxq->Mapping_bd);

    if (retStatus != 0) {
        DEBUG( (EFI_D_ERROR, "Iwl_Pcie_Rx_Alloc: DMA Map Failed for Mapping_bd,status = %d\n",retStatus ));
        goto error_bd;
    }

    //DMA Memory allocation for Driver Receiver status

    Status =  WirelessAllocatePage((VOID **)&rxq->rb_stts);

    if (EFI_ERROR(Status)) {
        DEBUG( (EFI_D_ERROR, "Iwl_Pcie_Rx_Alloc: rxq->rb_stts Memory Allocation Failed,status = %d\n",Status ));
        goto error_bd;
    }

    //Get virtual and Physical Address of the Page

    retStatus = Dma_Map(IwlTransContext,
                                 rxq->rb_stts,
                                 PAGE_SIZE,
                                 FROM_DEVICE,
                                 &rxq->rb_stts_dma,
                                 &rxq->Mapping_rb);

    if (retStatus != 0) {
        DEBUG( (EFI_D_ERROR, "Iwl_Pcie_Rx_Alloc: DMA Map Failed for rxq->Mapping_rb,status = %d\n",retStatus ));
        goto error_rb;
    }
    return 0;


error_rb:

    Status =  WirelessFreePage(rxq->rb_stts);

    if (EFI_ERROR(Status)) {
        DEBUG( (EFI_D_ERROR, "error_rb: Memory release failed , status = %d\n",Status ));
    }
    retStatus = Dma_Unmap(IwlTransContext,
                                   rxq->Mapping_bd);
    if (retStatus < 0)
    {
        DEBUG( (EFI_D_ERROR, "error_rb: Dma Unmap Failed,status = %d\n",retStatus ));
    }
error_bd:

    Status =  WirelessFreePage(rxq->bd);

    if (EFI_ERROR(Status)) {
        DEBUG( (EFI_D_ERROR, "error_bd: Memory release failed , status = %d\n",Status ));
    }
    rxq->bd_dma = 0;
    rxq->bd = NULL;
error:
    return -1;      //-ENOMEM;

}

VOID
Iwl_Pcie_Rxq_Free_Rbs(
        IWL_TRANS_CONTEXT *IwlTransContext
    )
{
    struct iwl_trans_pcie *trans_pcie = NULL;
    struct iwl_rxq *rxq = NULL;
    int i;
    EFI_STATUS Status;
    INT32 retStatus = 0;

    if(!IwlTransContext)
    {
        return;
    }
    trans_pcie = IWL_TRANS_GET_PCIE_TRANS(IwlTransContext);
    rxq = &trans_pcie->rxq;

    for (i = 0; i < RX_FREE_BUFFERS + RX_QUEUE_SIZE; i++) {
        if (!rxq->pool[i].page_dma)
            continue;
        retStatus =Dma_Unmap(IwlTransContext,
                                       rxq->pool[i].Mapping);
        if (retStatus < 0)
        {
            DEBUG( (EFI_D_ERROR, "Iwl_Pcie_Rxq_Free_Rbs:rxq Dma Unmap Failed ,status = %d\n",retStatus ));
        }
        /* Returning the page to the memory manager*/
        Status =  WirelessFreePage(rxq->pool[i].cpu_addr);
        if (EFI_ERROR(Status)) {
            DEBUG( (EFI_D_ERROR, "Iwl_Pcie_Rxq_Free_Rbs: rxq->cpuaddr Memory release failed,status = %d\n",Status ));
        }
        rxq->pool[i].cpu_addr=NULL;
        rxq->pool[i].page_dma=0;
    }
}

VOID
Iwl_Pcie_Rx_Init_Rxb_Lists(
    struct iwl_rxq *rxq
    )
{
    int i;
    if(rxq != NULL)
    {
        //List related calls needs to be implemented and reenabled here
        INIT_LIST_HEAD(&rxq->rx_free);
        INIT_LIST_HEAD(&rxq->rx_used);
        rxq->free_count = 0;

        for (i = 0; i < RX_FREE_BUFFERS + RX_QUEUE_SIZE; i++)
        {
            list_add(&rxq->pool[i].list, &rxq->rx_used);
        }
    }
}

VOID
Iwl_Pcie_Rx_Hw_Init(
        IWL_TRANS_CONTEXT *IwlTransContext,
    struct iwl_rxq *rxq
    )
{
    const u32 rfdnlog = RX_QUEUE_SIZE_LOG; /* 256 RBDs */

    if(!IwlTransContext || !rxq)
    {
        return;
    }
    /* Stop Rx DMA */
    iwl_write_direct32(IwlTransContext, FH_MEM_RCSR_CHNL0_CONFIG_REG, 0);
    // reset and flush pointers
    iwl_write_direct32(IwlTransContext, FH_MEM_RCSR_CHNL0_RBDCB_WPTR, 0);
    iwl_write_direct32(IwlTransContext, FH_MEM_RCSR_CHNL0_FLUSH_RB_REQ, 0);
    iwl_write_direct32(IwlTransContext, FH_RSCSR_CHNL0_RDPTR, 0);

    // Reset driver's Rx queue write index
    iwl_write_direct32(IwlTransContext, FH_RSCSR_CHNL0_RBDCB_WPTR_REG, 0);

    // Tell device where to find RBD circular buffer in DRAM
    iwl_write_direct32(IwlTransContext,FH_RSCSR_CHNL0_RBDCB_BASE_REG,(u32)(rxq->bd_dma >> 8));

    // Tell device where in DRAM to update its Rx status
    iwl_write_direct32(IwlTransContext, FH_RSCSR_CHNL0_STTS_WPTR_REG,
                       (u32)(rxq->rb_stts_dma >> 4));
    /**
     *   Enable Rx DMA
     *   FH_RCSR_CHNL0_RX_IGNORE_RXF_EMPTY is set because of HW bug in
     *   the credit mechanism in 5000 HW RX FIFO
     *   Direct rx interrupts to hosts
     *   Rx buffer size 4 or 8k
     *   RB timeout 0x10
     *   256 RBDs
     **/
    /// todo@ Hardcoded with FH_RCSR_RX_CONFIG_REG_VAL_RB_SIZE_4K value page size

    iwl_write_direct32(IwlTransContext, FH_MEM_RCSR_CHNL0_CONFIG_REG,
                       FH_RCSR_RX_CONFIG_CHNL_EN_ENABLE_VAL |
                       FH_RCSR_CHNL0_RX_IGNORE_RXF_EMPTY |
                       FH_RCSR_CHNL0_RX_CONFIG_IRQ_DEST_INT_HOST_VAL |
                       FH_RCSR_RX_CONFIG_REG_VAL_RB_SIZE_4K|
                       (RX_RB_TIMEOUT << FH_RCSR_RX_CONFIG_REG_IRQ_RBTH_POS)|
                       (rfdnlog << FH_RCSR_RX_CONFIG_RBDCB_SIZE_POS));

    /* Set interrupt coalescing timer to default (2048 usecs) */
    Iwl_Trans_Write8(IwlTransContext, CSR_INT_COALESCING, IWL_HOST_INT_TIMEOUT_DEF);

    if (IwlTransContext->IwlConfig->Device_Id == INTEL_7260_DEVICE_ID_1 || IwlTransContext->IwlConfig->Device_Id == INTEL_7260_DEVICE_ID_2 ||(IwlTransContext->IwlConfig->Device_Id == INTEL_3160_DEVICE_ID)) {
        /* W/A for interrupt coalescing bug in 7260 and 3160 */
        Iwl_Trans_Pcie_Set_Bits_Mask(IwlTransContext,
                                  CSR_INT_COALESCING,
                                  (UINT32) IWL_HOST_INT_OPER_MODE,
                                  (UINT32) IWL_HOST_INT_OPER_MODE);
    }

}

INT32
Iwl_Pcie_Rx_Init(
    IN IWL_TRANS_CONTEXT *IwlTransContext
    )
{
    struct iwl_trans_pcie *trans_pcie = IWL_TRANS_GET_PCIE_TRANS(IwlTransContext);
    struct iwl_rxq *rxq = &trans_pcie->rxq;
    int err,i;

    SetMem(&trans_pcie->rxq, sizeof(trans_pcie->rxq),0);

    if (!rxq->bd) {
        err = Iwl_Pcie_Rx_Alloc(IwlTransContext);
        if (err)
            return err;
    }

    Iwl_Pcie_Rxq_Free_Rbs(IwlTransContext);
    Iwl_Pcie_Rx_Init_Rxb_Lists(rxq);

    for (i = 0; i < RX_QUEUE_SIZE; i++)
        CopyMem(rxq->queue[i],NULL,sizeof(struct iwl_rx_mem_buffer));

    rxq->read = rxq->write = 0;
    rxq->write_actual = 0;
    SetMem(rxq->rb_stts, sizeof(*rxq->rb_stts),0);


    Iwl_Pcie_Rx_Replenish(IwlTransContext);


    Iwl_Pcie_Rx_Hw_Init(IwlTransContext, rxq);

    rxq->need_update = 1;
    Iwl_Pcie_Rxq_Inc_Wr_Ptr(IwlTransContext, rxq);

    return 0;
}


void
iwl_pcie_rx_free(
    struct iwl_trans_context *IwlTransContext
    )
{
    struct iwl_trans_pcie *trans_pcie = NULL;
    struct iwl_rxq *rxq = NULL;
    EFI_STATUS Status;
    INT32 retStatus = 0;

    if(!IwlTransContext)
    {
        return;
    }
    trans_pcie = IWL_TRANS_GET_PCIE_TRANS(IwlTransContext);
    rxq = &trans_pcie->rxq;
    /*if rxq->bd is NULL, it means that nothing has been allocated,
     * exit now */
    if (!rxq->bd) {
        DEBUG( (EFI_D_ERROR, "Free NULL rx context\n" ));
        return;
    }

    Iwl_Pcie_Rxq_Free_Rbs(IwlTransContext);

    retStatus = Dma_Unmap(
            IwlTransContext,
            rxq->Mapping_bd
            );
    if (retStatus < 0)
    {
        DEBUG( (EFI_D_ERROR, "iwl_pcie_rx_free:Mapping_bd Dma Unmap Failed ,status = %d\n",retStatus ));
    }

    Status =  WirelessFreePage(rxq->bd);

    if (EFI_ERROR(Status)) {
        DEBUG( (EFI_D_ERROR, "iwl_pcie_rx_free: rxq->bd Memory release failed,status = %d\n",Status ));
    }
    rxq->bd_dma = 0;
    rxq->bd = NULL;

    if (rxq->rb_stts) {
        retStatus = Dma_Unmap(
                IwlTransContext,
                rxq->Mapping_rb
                );
        if (retStatus < 0)
        {
            DEBUG( (EFI_D_ERROR, "iwl_pcie_rx_free:Mapping_rb Dma Unmap Failed ,status = %d\n",retStatus ));
        }

        Status =  WirelessFreePage(rxq->rb_stts);

        if (EFI_ERROR(Status)) {
            DEBUG( (EFI_D_ERROR, "iwl_pcie_rx_free: rxq->rb_stts Memory release failed,status = %d\n",Status ));
        }
    }
    rxq->rb_stts_dma = 0;
    rxq->rb_stts = NULL;

}
