/**
 * @file tx.c
 *
 * @details
 * This header file contains the code to implement transmit
 * operation for Intel UNDI driver.
 *
 * @copyright
 * Copyright (c) 2017 Dell Inc. All rights reserved. This
 * software and associated documentation (if any) is furnished
 * under a license and may only be used or copied in accordance
 * with the terms of the license. Except as permitted by such
 * license, no part of this software or documentation may be
 * reproduced, stored in a retrieval system, or transmitted in
 * any form or by any means without the express written consent
 * of Dell Inc.
 */

/******************************************************************************
 *
 * Copyright(c) 2003 - 2014 Intel Corporation. All rights reserved.
 *
 * Portions of this file are derived from the ipw3945 project, as well
 * as portions of the ieee80211 subsystem header files.
 *
 * This program is free software; you can redistribute it and/or modify it
 * under the terms of version 2 of the GNU General Public License as
 * published by the Free Software Foundation.
 *
 * This program is distributed in the hope that it will be useful, but WITHOUT
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
 * FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License for
 * more details.
 *
 * You should have received a copy of the GNU General Public License along with
 * this program; if not, write to the Free Software Foundation, Inc.,
 * 51 Franklin Street, Fifth Floor, Boston, MA 02110, USA
 *
 * The full GNU General Public License is included in this distribution in the
 * file called LICENSE.
 *
 * Contact Information:
 *  Intel Linux Wireless <ilw@linux.intel.com>
 * Intel Corporation, 5200 N.E. Elam Young Parkway, Hillsboro, OR 97124-6497
 *
 *****************************************************************************/
#include <Uefi.h>
#include <Library/DebugLib.h>
#include "iwl-prph.h"
#include "internal.h"
#include "fw-api-Tx.h"
#include "IwlTransMain.h"
#include "IwlMvmOps.h"
#include "WirelessMemoryManager.h"

enum dma_data_direction {
    DMA_BIDIRECTIONAL = 0,
    DMA_FROM_DEVICE,
    DMA_TO_DEVICE,
    DMA_NONE,
};

enum txq_tbs
{
    TB0,
    TB1,
    TB2,
    TB3
};

void Iwl_Force_Nmi(struct iwl_trans_context *IwlTransContext);

#define IWL_TX_CRC_SIZE 4
#define IWL_TX_DELIMITER_SIZE 4
#define offsetof_Iwl(TYPE, MEMBER) ((size_t) &((TYPE *)0)->MEMBER)
#define MAX_TXRESP_PKT_SIZE 1000

#define min_t(type, x, y)  ((x) < (y) ? (x) : (y))

/** 1: Driver inserted 2 bytes pad after the MAC header, for (required) dword
 *    alignment of frame's payload data field.
 * 0: No pad
 * Set this for MAC headers with 26 or 30 bytes, i.e. those with QOS or ADDR4
 * field (but not both).  Driver must align frame data (i.e. data following
 * MAC header) to DWORD boundary.
 **/

#define TX_CMD_FLG_MH_PAD_MSK CONV_TO_LE32_BIT(1 << 20)

static dma_addr_t
iwl_pcie_get_scratchbuf_dma(
    struct iwl_txq *txq,
    int idx
    )
{
    if(txq != NULL)
    {
        return (txq->scratchbufs_dma + (sizeof(struct iwl_pcie_txq_scratch_buf) * idx));
    }
    else
    {
        return 0;
    }
}

UINTN Check_num_of_pages(UINTN len)
{
    int num_page;
    if (len <= PAGE_SIZE)
        num_page = 1;
    else {
        num_page = (UINT16)(len / PAGE_SIZE);
        if (len % PAGE_SIZE)
            num_page++;
    }
    return num_page;
}

dma_addr_t dma_map_single(
    IWL_TRANS_CONTEXT *IwlTransContext,
    INTN txq_id,
    UINT8 tbs_idx,
    void *ptr,
    size_t size,
    enum dma_data_direction dir
    )
{
    INT32 retStatus = 0;
    INT32 tfd_idx;
    dma_addr_t phy_addr = 0x0;
    struct iwl_trans_pcie *trans_pcie = NULL;
    struct iwl_txq *txq = NULL;
    struct iwl_queue *q = NULL;

    if (!ptr || !IwlTransContext)
    {
        return 0;
    }
    trans_pcie = IWL_TRANS_GET_PCIE_TRANS(IwlTransContext);
    txq = &trans_pcie->txq[txq_id];
    q = &txq->q;
    tfd_idx = get_cmd_index(q, q->write_ptr);

    retStatus = Dma_Map(IwlTransContext,
                        ptr,
                       (UINT32) size,
                        dir,
                        &phy_addr,
                        &txq->tbs_mapping[tfd_idx][tbs_idx]
                        );

    if (retStatus != 0) {
        DEBUG( (EFI_D_ERROR, "dma_map_single: DMA Map Failed for tbs_mapping,status = %d\n",retStatus ));
        return 0;
    }

    return phy_addr;

}


/*************** DMA-QUEUE-GENERAL-FUNCTIONS  *****
 * DMA services
 *
 * Theory of operation
 *
 * A Tx or Rx queue resides in host DRAM, and is comprised of a circular buffer
 * of buffer descriptors, each of which points to one or more data buffers for
 * the device to read from or fill.  Driver and device exchange status of each
 * queue via "read" and "write" pointers.  Driver keeps minimum of 2 empty
 * entries in each circular buffer, to protect against confusing empty and full
 * queue states.
 *
 * The device reads or writes the data in the queues via the device's several
 * DMA/FIFO channels.  Each queue is mapped to a single DMA channel.
 *
 * For Tx queue, there are low mark and high mark limits. If, after queuing
 * the packet for Tx, free space become < low mark, Tx queue stopped. When
 * reclaiming packets (on 'tx done IRQ), if free space become > high mark,
 * Tx queue resumed.
 *
 ***************************************************/
static
int
iwl_queue_space(
    const struct iwl_queue *q
    )
{

    unsigned int max;
    unsigned int used;

    if(!q)
    {
        return -1;
    }
    /*
     * To avoid ambiguity between empty and completely full queues, there
     * should always be less than q->n_bd elements in the queue.
     * If q->n_window is smaller than q->n_bd, there is no need to reserve
     * any queue entries for this purpose.
     */
    if (q->n_window < q->n_bd)
        max = q->n_window;
    else
        max = q->n_bd - 1;

    /*
     * q->n_bd is a power of 2, so the following is equivalent to modulo by
     * q->n_bd and is well defined for negative dividends.
     */
    used = (q->write_ptr - q->read_ptr) & (q->n_bd - 1);

    if (used > max)
        return 0;

    return max - used;
}

/*
 * iwl_queue_init - Initialize queue's high/low-water and read/write indexes
 */
static
int
iwl_queue_init(
    struct iwl_queue *q,
    int count,
    int slots_num,
    u32 id
    )
{
    if(q != NULL)
    {
        q->n_bd = count;
        q->n_window = slots_num;
        q->id = id;

        q->low_mark = q->n_window / 4;
        if (q->low_mark < 4)
            q->low_mark = 4;

        q->high_mark = q->n_window / 8;
        if (q->high_mark < 2)
            q->high_mark = 2;

        q->write_ptr = 0;
        q->read_ptr = 0;
        return 0;
    }
    else
    {
        return -1;
    }
}

static
void
iwl_pcie_free_dma_ptr(
    struct iwl_trans_context *IwlTransContext,
    struct iwl_dma_ptr *ptr
    )
{
    EFI_STATUS Status;
    INT32 Ret = 0;

    if (!IwlTransContext || !ptr || !ptr->addr)
    {
        return;
    }

    Ret = Dma_Unmap(IwlTransContext, ptr->mapping);
    if (Ret < 0)
    {
        DEBUG( (EFI_D_ERROR, "iwl_pcie_free_dma_ptr:DMA UnMap Failed,Status :%d\n",Ret ));
    }

    Status = IwlTransContext->PciIoFuncs->FreeBuffer(IwlTransContext->PciIoFuncs,
                                                     Check_num_of_pages(ptr->size),
                                                     ptr->addr
                                                     );


    if (EFI_ERROR(Status)) {
        DEBUG( (EFI_D_ERROR, "iwl_pcie_free_dma_ptr: Memory release failed,status = %d\n",Status ));
    }
}

/*
 * iwl_pcie_txq_update_byte_cnt_tbl - Set up entry in Tx byte-count array
 */
static
void
iwl_pcie_txq_update_byte_cnt_tbl(
    struct iwl_trans_context *IwlTransContext,
    struct iwl_txq *txq,
    u16 byte_cnt
    )
{
    struct iwlagn_scd_bc_tbl *scd_bc_tbl;
    struct iwl_trans_pcie *trans_pcie = NULL;
    int write_ptr = 0;
    int txq_id = 0;
    u8 sec_ctl = 0;
    u8 sta_id = 0;
    u16 len = byte_cnt + IWL_TX_CRC_SIZE + IWL_TX_DELIMITER_SIZE;
    __le16 bc_ent;
    INT32 retStatus = 0;
    struct iwl_tx_cmd *tx_cmd = NULL;

    if(!IwlTransContext || !txq)
    {
        return;
    }
    trans_pcie = IWL_TRANS_GET_PCIE_TRANS(IwlTransContext);
    write_ptr = txq->q.write_ptr;
    txq_id = txq->q.id;
    tx_cmd = (void *) txq->entries[txq->q.write_ptr].cmd->Payload;
    scd_bc_tbl = trans_pcie->scd_bc_tbls.addr;

    if((len > 0xFFF || write_ptr >= TFD_QUEUE_SIZE_MAX))
        DEBUG( (EFI_D_ERROR, "write_ptr exceeds TFD QUEUE SIZE MAX \r\n" ));

    sta_id = tx_cmd->sta_id;
    sec_ctl = tx_cmd->sec_ctl;

    switch (sec_ctl & TX_CMD_SEC_MSK) {
        case TX_CMD_SEC_CCM:
            len += WIFI_CCMP_MIC_LENGTH;
            break;
        case TX_CMD_SEC_TKIP:
            len += WIFI_TKIP_ICV_LENGTH;
            break;
        case TX_CMD_SEC_WEP:
            len += WIFI_WEP_IV_LENGTH + WIFI_WEP_ICV_LENGTH;
            break;
    }

    if (trans_pcie->bc_table_dword)
        len = DIV_ROUND_UP(len, 4);

    bc_ent = CONV_TO_LE16_BIT(len | (sta_id << 12));

    retStatus = Dma_Unmap(IwlTransContext, trans_pcie->scd_bc_tbls.mapping);

    if (retStatus < 0)
    {
        DEBUG( (EFI_D_ERROR, "iwl_pcie_txq_update_byte_cnt_tbl:DMA Un Map Failed,Status : %d\n",retStatus ));
    }

    scd_bc_tbl[txq_id].tfd_offset[write_ptr] = bc_ent;

    if (write_ptr < TFD_QUEUE_SIZE_BC_DUP)
        scd_bc_tbl[txq_id].
        tfd_offset[TFD_QUEUE_SIZE_MAX + write_ptr] = bc_ent;

    retStatus = Dma_Map(IwlTransContext,
                     trans_pcie->scd_bc_tbls.addr,
                     (UINT32) trans_pcie->scd_bc_tbls.size,
                     TO_DEVICE,
                     (UINT64*)&trans_pcie->scd_bc_tbls.dma,
                     &trans_pcie->scd_bc_tbls.mapping);

    if (retStatus < 0)
    {
        DEBUG( (EFI_D_ERROR, "iwl_pcie_txq_update_byte_cnt_tbl: DMA Map failed,status = %d\n",retStatus ));
    }
}

static
void
iwl_pcie_txq_inval_byte_cnt_tbl(
    struct iwl_trans_context *IwlTransContext,
    struct iwl_txq *txq
    )
{

    struct iwl_trans_pcie *trans_pcie = NULL;
    struct iwlagn_scd_bc_tbl *scd_bc_tbl = NULL;
    int txq_id = 0;
    int read_ptr = 0;
    u8 sta_id = 0;
    __le16 bc_ent;
    INT32 retStatus = 0;
    struct iwl_tx_cmd *tx_cmd = NULL;

    if(!IwlTransContext || !txq)
    {
        return;
    }

    trans_pcie = IWL_TRANS_GET_PCIE_TRANS(IwlTransContext);
    scd_bc_tbl = trans_pcie->scd_bc_tbls.addr;
    txq_id = txq->q.id;
    read_ptr = txq->q.read_ptr;
    tx_cmd = (void *)txq->entries[txq->q.read_ptr].cmd->Payload;

    if((read_ptr >= TFD_QUEUE_SIZE_MAX))
        DEBUG( (EFI_D_INFO, "read_ptr exceeds TFD QUEUE SIZE MAX \r\n" ));

    if (txq_id != trans_pcie->Cmd_Queue)
        sta_id = tx_cmd->sta_id;

    bc_ent = CONV_TO_LE16_BIT(1 | (sta_id << 12));

    retStatus = Dma_Unmap(IwlTransContext, trans_pcie->scd_bc_tbls.mapping);

    if (retStatus < 0)
    {
        DEBUG( (EFI_D_ERROR, "iwl_pcie_txq_inval_byte_cnt_tbl: DMA Un Map Failed,status = %d\n",retStatus ));
    }

    scd_bc_tbl[txq_id].tfd_offset[read_ptr] = bc_ent;

    if (read_ptr < TFD_QUEUE_SIZE_BC_DUP)
        scd_bc_tbl[txq_id].
        tfd_offset[TFD_QUEUE_SIZE_MAX + read_ptr] = bc_ent;

    retStatus = Dma_Map(IwlTransContext,
                     trans_pcie->scd_bc_tbls.addr,
                     (UINT32) trans_pcie->scd_bc_tbls.size,
                     TO_DEVICE,
                     (UINT64*)(UINTN) &trans_pcie->scd_bc_tbls.dma,
                     &trans_pcie->scd_bc_tbls.mapping);

    if (retStatus != 0) {
        DEBUG( (EFI_D_ERROR, "iwl_pcie_txq_inval_byte_cnt_tbl:DMA Map Failed,status = %d\n",retStatus ));
    }
}

/*
 * iwl_pcie_txq_inc_wr_ptr - Send new write index to hardware
 */
void
iwl_pcie_txq_inc_wr_ptr(
    struct iwl_trans_context *IwlTransContext,
    struct iwl_txq *txq
    )
{
    struct iwl_trans_pcie *trans_pcie = NULL;
    u32 reg = 0;
    int txq_id = 0;

    if(!IwlTransContext || !txq)
    {
        return;
    }
    trans_pcie = IWL_TRANS_GET_PCIE_TRANS(IwlTransContext);
    txq_id = txq->q.id;

    if (txq->need_update == 0)
        return;

    /*
     * explicitly wake up the NIC if:
     * 1. shadow registers aren't enabled
     * 2. NIC is woken up for CMD regardless of shadow outside this function
     * 3. there is a chance that the NIC is asleep
     */
    if (!IwlTransContext->IwlConfig->Base_Params->Shadow_Reg_Enable &&
        txq_id != trans_pcie->Cmd_Queue &&
        test_bit(STATUS_TPOWER_PMI, &IwlTransContext->Status)) {
        /*
         * wake up nic if it's powered down ...
         * uCode will wake up, and interrupt us again, so next
         * time we'll skip this part.
         */
        reg = Iwl_Trans_Pcie_Read32(IwlTransContext, CSR_UCODE_DRV_GP1);

        if (reg & CSR_UCODE_DRV_GP1_BIT_MAC_SLEEP) {
            DEBUG( (EFI_D_INFO, "iwl_pcie_txq_inc_wr_ptr: Tx queue %d requesting wakeup, GP1 = 0x%x\n",
                                                  txq_id, reg));
            Iwl_Trans_Pcie_Set_Bits_Mask(IwlTransContext, CSR_GP_CNTRL,
                                          CSR_GP_CNTRL_REG_FLAG_MAC_ACCESS_REQ,
                                          CSR_GP_CNTRL_REG_FLAG_MAC_ACCESS_REQ);
            return;
        }
    }

    DEBUG( (EFI_D_INFO, "iwl_pcie_txq_inc_wr_ptr: Q:%d WR: 0x%x\n", txq_id, txq->q.write_ptr ));

    Iwl_Trans_Pcie_Write32(IwlTransContext, HBUS_TARG_WRPTR, txq->q.write_ptr | (txq_id << 8));

    txq->need_update = 0;

}

//Added below apis from linux kernel base code.
static
UINT32
__get_unaligned_le32(
    const u8 *p
    )
{
    return p[0] | p[1] << 8 | p[2] << 16 | p[3] << 24;
}

static
UINT32
get_unaligned_le32(
    const void *p
    )
{
    return __get_unaligned_le32((const u8 *)p);
}

static
dma_addr_t
iwl_pcie_tfd_tb_get_addr(
    struct iwl_tfd *tfd,
    u8 idx
    )
{

    struct iwl_tfd_tb *tb = &tfd->tbs[idx];

    dma_addr_t addr = get_unaligned_le32(&tb->lo);
    if (sizeof(dma_addr_t) > sizeof(u32))
        addr |=
            ((dma_addr_t)(le16_to_cpu(tb->hi_n_len) & 0xF) << 16) << 16;

    return addr;
}

static
u16
iwl_pcie_tfd_tb_get_len(
    struct iwl_tfd *tfd,
    u8 idx
    )
{
    struct iwl_tfd_tb *tb = &tfd->tbs[idx];

    return le16_to_cpu(tb->hi_n_len) >> 4;
}

static
void
__put_unaligned_le16(
    u32 val,
    u8 *p
    )
{
    if(p != NULL)
    {
        *p++ = (u8)val;
        *p++ = (u8)(val >> 8);
    }
}

static
void
__put_unaligned_le32(
    u64 val,
    u8 *p
    )
{
    __put_unaligned_le16((u32)(val >> 16), p + 2);
    __put_unaligned_le16((u32)val, p);
}

static
void
put_unaligned_le32(
    u64 val,
    void *p
    )
{
    __put_unaligned_le32(val, p);
}

static
void
iwl_pcie_tfd_set_tb(
    struct iwl_tfd *tfd,
    u8 idx,
    dma_addr_t addr,
    u16 len
    )
{

    struct iwl_tfd_tb *tb = NULL;
    u16 hi_n_len = len << 4;

    if(tfd != NULL)
    {
        tb = &tfd->tbs[idx];
        put_unaligned_le32(addr, &tb->lo);
        if (sizeof(dma_addr_t) > sizeof(u32))
            hi_n_len |= ((addr >> 16) >> 16) & 0xF;

        tb->hi_n_len = CONV_TO_LE16_BIT(hi_n_len);

        tfd->num_tbs = idx + 1;
    }

}

static
u8
iwl_pcie_tfd_get_num_tbs(
    struct iwl_tfd *tfd
    )
{
    return tfd->num_tbs & 0x1f;
}

static
void
iwl_pcie_tfd_unmap_and_free(
    struct iwl_trans_context *IwlTransContext,
    struct iwl_tfd *tfd,
    struct iwl_txq *txq
    )
{

    int num_tbs;
    int rd_ptr = 0;
    int idx = 0,i;
    EFI_STATUS Status;
    INT32 retStatus = 0;

    if(!IwlTransContext || !tfd || !txq)
    {
        return;
    }
    /* rd_ptr is bounded by n_bd and idx is bounded by n_window */
    rd_ptr = txq->q.read_ptr;

    idx = get_cmd_index(&txq->q, rd_ptr);

    /* Sanity check on number of chunks */
    num_tbs = iwl_pcie_tfd_get_num_tbs(tfd);
    if (num_tbs > IWL_NUM_OF_TBS) {
        DEBUG( (EFI_D_ERROR, "iwl_pcie_tfd_unmap: Too many chunks: %d\n", num_tbs ));

        return;
    }
    /*
     * first TB is never freed - it's the scratchbuf data.
     * If you free this, fw will be crashed at some point of data
     * transmission.
     * Reason: During data transmission, first 16 bytes
     * are copied to scratch buffer. Memory allocation for scratch buffer
     * done during rx init time. In Tx we are not allocating memory for
     * scratch buffer and we are using allocated memory.
     * If we free this scratch buffer, once 256 TFD's completed in Tx,
     * fw will be crashed.
     */
    /*Here the Index zero points to the tb which holds the excess
     * command header during command transmission
     * */

    for(i = 0; i < IWL_NUM_OF_TBS; i++)

        if(txq->entries[idx].tb_cpuaddr[i] != NULL) {
            retStatus = Dma_Unmap(IwlTransContext,
                                   txq->tbs_mapping[idx][i]);

            if (retStatus < 0)
            {
                DEBUG( (EFI_D_ERROR, "iwl_pcie_tfd_unmap: tb unmap failed,status = %d\n",retStatus ));
            }

            Status =  WirelessFreePage((VOID**)txq->entries[idx].tb_cpuaddr[i]);

            if (EFI_ERROR(Status))
            {
                DEBUG( (EFI_D_ERROR, "iwl_pcie_tfd_unmap: tb_cpuaddr Memory release failed,status = %d\n",Status ));
            }
            txq->entries[idx].tb_cpuaddr[i] = NULL;
            txq->entries[idx].tb_phys[i] = 0;
        }
    tfd->num_tbs = 0;

    return;
}

/*
 * iwl_pcie_txq_free_tfd - Free all chunks referenced by TFD [txq->q.read_ptr]
 * @IwlTransContext - transport private data
 * @txq - tx queue
 * @dma_dir - the direction of the DMA mapping
 *
 * Does NOT advance any TFD circular buffer read/write indexes
 * Does NOT free the TFD itself (which is within circular buffer)
 */
static
void
iwl_pcie_txq_free_tfd(
    struct iwl_trans_context *IwlTransContext,
    struct iwl_txq *txq
    )
{

    struct iwl_tfd *tfd_tmp = NULL;
    int rd_ptr = 0;

    if(IwlTransContext != NULL && txq !=  NULL)
    {
        tfd_tmp = txq->tfds;
        /* rd_ptr is bounded by n_bd and idx is bounded by n_window */
        rd_ptr = txq->q.read_ptr;
        /* We have only q->n_window txq->entries, but we use q->n_bd tfds */
        iwl_pcie_tfd_unmap_and_free(IwlTransContext, &tfd_tmp[rd_ptr], txq);
    }
}

static
int
iwl_pcie_txq_build_tfd(
    struct iwl_trans_context *IwlTransContext,
    struct iwl_txq *txq,
    dma_addr_t addr,
    u16 len,
    u8 reset
    )
{

    struct iwl_queue *q;
    struct iwl_tfd *tfd, *tfd_tmp;
    u32 num_tbs;

    if(!txq)
    {
        return -EINVAL;
    }
    q = &txq->q;
    tfd_tmp = txq->tfds;
    tfd = &tfd_tmp[q->write_ptr];

    if (reset)
        SetMem(tfd, sizeof(*tfd),0);

    num_tbs = iwl_pcie_tfd_get_num_tbs(tfd);

    /* Each TFD can point to a maximum 20 Tx buffers */
    if (num_tbs > IWL_NUM_OF_TBS) {
        DEBUG( (EFI_D_ERROR,"iwl_pcie_txq_build_tfd: Error can not send more than %d chunks\n",
                                               IWL_NUM_OF_TBS));

        return -EINVAL;
    }

    if ((addr & ~IWL_TX_DMA_MASK)) {
        DEBUG( (EFI_D_ERROR, "iwl_pcie_txq_build_tfd:Unaligned address = %llx\n",(unsigned long long)addr ));
        return -EINVAL;
    }


    iwl_pcie_tfd_set_tb(tfd, (u8)num_tbs, addr, len);

    return 0;
}

static
int
iwl_pcie_txq_alloc(
    struct iwl_trans_context *IwlTransContext,
    struct iwl_txq *txq,
    int slots_num,
    u32 txq_id
    )
{
    struct iwl_trans_pcie *trans_pcie = NULL;
    size_t tfd_sz = sizeof(struct iwl_tfd) * TFD_QUEUE_SIZE_MAX;
    size_t scratchbuf_sz;
    int i,j;
    EFI_STATUS Status;
    INT32 retStatus = 0;

    if(!IwlTransContext || !txq)
    {
        return -1;
    }

    trans_pcie = IWL_TRANS_GET_PCIE_TRANS(IwlTransContext);

    for (i = 0; i < TFD_QUEUE_SIZE_MAX; i++) {
        for(j = 0; i < IWL_NUM_OF_TBS; i++)
            txq->tbs_mapping[i][j] = NULL;
    }
    txq->tfds_mapping = NULL;
    txq->scratch_buf_mapping = NULL;
    txq->trans_pcie = trans_pcie;

    txq->q.n_window = slots_num;


    txq->entries = WifiAllocatepool(slots_num * (sizeof(struct iwl_pcie_txq_entry)),__FUNCTION__);
    if(txq->entries == NULL) {
        DEBUG( (EFI_D_ERROR, "iwl_pcie_txq_alloc: Txq Entry: Allocate pool failed\n" ));

        goto error;
    }

    if (txq_id == trans_pcie->Cmd_Queue)
        for (i = 0; i < slots_num; i++) {

            txq->entries[i].cmd = WifiAllocatepool(sizeof(struct iwl_device_cmd),__FUNCTION__);
            if(txq->entries[i].cmd == NULL) {
                DEBUG( (EFI_D_ERROR, "iwl_pcie_txq_alloc:Txq Entry-> Device Command: Allocate pool failed\n" ));
                goto error;
            }
        }

    /* Circular buffer of transmit frame descriptors (TFDs),
     * shared with device */
    //DMA Malloc Here

    Status = (INT32)IwlTransContext->PciIoFuncs->AllocateBuffer(IwlTransContext->PciIoFuncs,
                                                                AllocateAnyPages,
                                                                EfiBootServicesData,
                                                                Check_num_of_pages(tfd_sz),
                                                                (void **)&txq->tfds,
                                                                0);



    if (EFI_ERROR(Status)) {
        DEBUG( (EFI_D_ERROR, "iwl_pcie_txq_alloc:No memory can be allocated for tfds,status = %d\n",Status ));
        goto error;
    }

    retStatus = Dma_Map(IwlTransContext,
                     txq->tfds,
                     (UINT32) tfd_sz,
                     TO_AND_FROM_DEVICE,
                     (UINT64*)(UINTN) &txq->q.dma_addr,
                     &txq->tfds_mapping
                     );

    if (retStatus != 0) {
        DEBUG( (EFI_D_ERROR, "iwl_pcie_txq_alloc:DMA Map Failed for tfds_mapping,status = %d\n",retStatus ));
        goto error;
    }

    scratchbuf_sz = sizeof(*txq->scratchbufs) * slots_num;


    Status = (INT32)IwlTransContext->PciIoFuncs->AllocateBuffer(IwlTransContext->PciIoFuncs,
                                                                AllocateAnyPages,
                                                                EfiBootServicesData,
                                                                Check_num_of_pages(scratchbuf_sz),
                                                                (void **)&txq->scratchbufs,
                                                                0);



    if (EFI_ERROR(Status)) {
        DEBUG( (EFI_D_ERROR, "iwl_pcie_txq_alloc:No memory can be allocated for scratchbufs,status = %d\n",Status ));
        goto err_free_tfds;
    }

    SetMem(txq->scratchbufs, sizeof(*txq->scratchbufs),0);

    retStatus = Dma_Map(IwlTransContext,
                     txq->scratchbufs,
                     (UINT32) scratchbuf_sz,
                     TO_AND_FROM_DEVICE,
                     (UINT64*)(UINTN) &txq->scratchbufs_dma,
                     &txq->scratch_buf_mapping);

    if (retStatus != 0) {
        DEBUG( (EFI_D_ERROR, "iwl_pcie_txq_alloc:DMA Map Failed for scratch_buf_mapping,status = %d\n",retStatus ));
        goto err_free_tfds;
    }

    txq->q.id = txq_id;

    return 0;

err_free_tfds:
    retStatus = Dma_Unmap(IwlTransContext, txq->tfds_mapping);

    if (retStatus < 0)
    {
        DEBUG( (EFI_D_ERROR, "iwl_pcie_txq_alloc: tfds_mapping unmap failed,status = %d\n",retStatus ));
    }


    Status = (INT32)IwlTransContext->PciIoFuncs->FreeBuffer(IwlTransContext->PciIoFuncs,
                                                            Check_num_of_pages(tfd_sz),
                                                            txq->tfds);


    if (EFI_ERROR(Status)) {
        DEBUG( (EFI_D_ERROR, "iwl_pcie_txq_alloc: tfds Memory release failed\n" ));
    }
error:
    if (txq->entries && txq_id == trans_pcie->Cmd_Queue)
    {
        for (i = 0; i < slots_num; i++)
        {
            if (txq->entries[i].cmd != NULL)
            {
                Status = (INT32)WifiFreePool(txq->entries[i].cmd,__FUNCTION__,__LINE__);
                txq->entries[i].cmd = NULL;
            }
        }
    }
    if (txq->entries != NULL)
    {
        WifiFreePool(txq->entries,__FUNCTION__,__LINE__);
        txq->entries = NULL;
    }
    return -ENOMEM;
}

static
int
iwl_pcie_txq_init(
    struct iwl_trans_context *IwlTransContext,
    struct iwl_txq *txq,
    int slots_num,
    u32 txq_id
    )
{

    int ret;

    if(!txq)
    {
        return -1;
    }
    txq->need_update = 0;

    /* TFD_QUEUE_SIZE_MAX must be power-of-two size, otherwise
     * iwl_queue_inc_wrap and iwl_queue_dec_wrap are broken. */

    /* Initialize queue's high/low-water marks, and head/tail indexes */
    ret = iwl_queue_init(&txq->q, TFD_QUEUE_SIZE_MAX, slots_num,txq_id);

    if (ret)
        return ret;

    /*
     * Tell nic where to find circular buffer of Tx Frame Descriptors for
     * given Tx queue, and enable the DMA channel used for that queue.
     * Circular buffer (TFD queue in DRAM) physical base address */
    iwl_write_direct32(IwlTransContext, FH_MEM_CBBC_QUEUE(txq_id),(u32)(txq->q.dma_addr >> 8));

    return 0;


}

/*
 * iwl_pcie_txq_unmap -  Unmap any remaining DMA mappings and free skb's
 */
static
void
iwl_pcie_txq_unmap(
        struct iwl_trans_context *IwlTransContext,
    int txq_id
    )
{
    struct iwl_trans_pcie *trans_pcie = NULL;
    struct iwl_txq *txq = NULL;
    struct iwl_queue *q = NULL;

    if(!IwlTransContext)
    {
        return;
    }

    trans_pcie = IWL_TRANS_GET_PCIE_TRANS(IwlTransContext);
    txq = &trans_pcie->txq[txq_id];
    q = &txq->q;

    if (!q->n_bd)
        return;

    while (q->write_ptr != q->read_ptr) {
        iwl_pcie_txq_free_tfd(IwlTransContext, txq);
        q->read_ptr = iwl_queue_inc_wrap(q->read_ptr, q->n_bd);
    }

    txq->active = false;
}

/*
 * iwl_pcie_txq_free - Deallocate DMA queue.
 * @txq: Transmit queue to deallocate.
 *
 * Empty queue by removing and destroying all BD's.
 * Free all buffers.
 * 0-fill, but do not free "txq" descriptor structure.
 */
static
void
iwl_pcie_txq_free(
        struct iwl_trans_context *IwlTransContext,
    int txq_id
    )
{
    struct iwl_trans_pcie *trans_pcie = NULL;
    struct iwl_txq *txq = NULL;
    int i;

    size_t tfd_sz = sizeof(struct iwl_tfd) * TFD_QUEUE_SIZE_MAX;
    size_t scratchbuf_sz;

    EFI_STATUS Status;
    INT32 retStatus = 0;
    u16 num_page;
    INT32 Ret = 0;

    if(!IwlTransContext)
    {
        return;
    }
    trans_pcie = IWL_TRANS_GET_PCIE_TRANS(IwlTransContext);
    txq = &trans_pcie->txq[txq_id];
    //DeadCode
/*
    if (!txq)
        return;*/

    iwl_pcie_txq_unmap(IwlTransContext, txq_id);

    /* De-alloc array of command/tx buffers */
    if (txq->entries && txq_id == trans_pcie->Cmd_Queue)
        for (i = 0; i < txq->q.n_window; i++) {
            if(txq->entries[i].cmd != NULL) {
                WifiFreePool(txq->entries[i].cmd,__FUNCTION__,__LINE__);
                txq->entries[i].cmd = NULL;
            }
            if(txq->entries[i].free_buf != NULL) {
                WifiFreePool((void *)txq->entries[i].free_buf,__FUNCTION__,__LINE__);
                txq->entries[i].free_buf = NULL;
            }
        }



    /* De-alloc circular buffer of TFDs */
    if (txq->q.n_bd) {
        retStatus = Dma_Unmap(IwlTransContext, txq->tfds_mapping);

        if (retStatus < 0)
        {
            DEBUG( (EFI_D_ERROR, "iwl_pcie_txq_free: tfds_mapping unmap failed,status = %d\n",retStatus ));
        }
        if (tfd_sz < PAGE_SIZE)
        {
            num_page = 1;
        }
        else
        {
            num_page = (UINT16)(tfd_sz / PAGE_SIZE);
            Ret = tfd_sz % PAGE_SIZE;
            if (Ret)
                num_page++;
        }

        Status = (INT32)IwlTransContext->PciIoFuncs->FreeBuffer(IwlTransContext->PciIoFuncs,
                                                                num_page,
                                                                txq->tfds);



        if (EFI_ERROR(Status)) {
            DEBUG( (EFI_D_ERROR, "iwl_pcie_txq_free: tfds Memory release failed,status = %d\n",Status ));
        }
        txq->q.dma_addr = 0;
        retStatus = Dma_Unmap(IwlTransContext, txq->scratch_buf_mapping);
        if (retStatus < 0)
        {
            DEBUG( (EFI_D_ERROR, "iwl_pcie_txq_free: scratch_buf_mapping unmap failed,status = %d\n",retStatus ));
        }

        scratchbuf_sz = sizeof(*txq->scratchbufs) * (txq->q.n_window);

        if (scratchbuf_sz < PAGE_SIZE)
            num_page = 1;
        else {
            num_page = (UINT16)(scratchbuf_sz / PAGE_SIZE);
            Ret = scratchbuf_sz % PAGE_SIZE;
            if (Ret)
                num_page++;
        }

        Status = (INT32)IwlTransContext->PciIoFuncs->FreeBuffer(IwlTransContext->PciIoFuncs,
                                                                num_page,
                                                                txq->scratchbufs);



        if (EFI_ERROR(Status)) {
            DEBUG( (EFI_D_ERROR, "iwl_pcie_txq_free: scratchbufs Memory release failed,status = %d\n",Status ));
        }
    }
    if(txq->entries != NULL) {
        WifiFreePool(txq->entries,__FUNCTION__,__LINE__);
        txq->entries = NULL;
    }

    /* 0-fill queue descriptor structure */
    SetMem(txq,sizeof(*txq),0);
}

/*
 * Activate/Deactivate Tx DMA/FIFO channels according tx fifos mask
 */

static
void
iwl_pcie_txq_set_sched(
    struct iwl_trans_context *IwlTransContext,
    u32 mask
    )
{

    IwlWritePrph(IwlTransContext, SCD_TXFACT, mask);

}

void
iwl_pcie_tx_start(
    struct iwl_trans_context *IwlTransContext,
    u32 scd_base_addr
    )
{

    struct iwl_trans_pcie *trans_pcie = NULL;
    int nq = 0;
    int chan;
    u32 reg_val;
    int clear_dwords = (SCD_TRANS_TBL_OFFSET_QUEUE(nq) -
                        SCD_CONTEXT_MEM_LOWER_BOUND) / sizeof(u32);


    if(!IwlTransContext)
    {
        return;
    }
    trans_pcie = IWL_TRANS_GET_PCIE_TRANS(IwlTransContext);
    nq = IwlTransContext->IwlConfig->Base_Params->Num_Of_Queues;

    trans_pcie->Scd_Base_Addr =
        IwlReadPrph(IwlTransContext, SCD_SRAM_BASE_ADDR);

    if(scd_base_addr != 0 && scd_base_addr != trans_pcie->Scd_Base_Addr) {

        DEBUG( (EFI_D_INFO, "iwl_pcie_tx_start: E=WARN ON \r\n" ));
    }


    /* reset context data, TX status and translation data */
    Iwl_Trans_Pcie_Write_Mem(IwlTransContext, trans_pcie->Scd_Base_Addr +
                        SCD_CONTEXT_MEM_LOWER_BOUND,
                        NULL, clear_dwords);

    IwlWritePrph(IwlTransContext, SCD_DRAM_BASE_ADDR,(u32)(trans_pcie->scd_bc_tbls.dma >> 10));

    /* The chain extension of the SCD doesn't work well. This feature is
     * enabled by default by the HW, so we need to disable it manually.
     */
    IwlWritePrph(IwlTransContext, SCD_CHAINEXT_EN, 0);

    Iwl_Trans_Pcie_Txq_Enable(IwlTransContext, trans_pcie->Cmd_Queue,
                            trans_pcie->cmd_fifo,
                            -1,
                            IWL_MAX_TID_COUNT,
                            IWL_FRAME_LIMIT,
                            0);

    /* Activate all Tx DMA/FIFO channels */
    iwl_pcie_txq_set_sched(IwlTransContext, IWL_MASK(0, 7));

    /* Enable DMA channel */
    for (chan = 0; chan < FH_TCSR_CHNL_NUM; chan++)
        iwl_write_direct32(IwlTransContext, FH_TCSR_CHNL_TX_CONFIG_REG(chan),
                           FH_TCSR_TX_CONFIG_REG_VAL_DMA_CHNL_ENABLE |
                           FH_TCSR_TX_CONFIG_REG_VAL_DMA_CREDIT_ENABLE);

    /* Update FH chicken bits */
    reg_val = IwlReadDirect32(IwlTransContext, FH_TX_CHICKEN_BITS_REG);
    iwl_write_direct32(IwlTransContext, FH_TX_CHICKEN_BITS_REG,
                       reg_val | FH_TX_CHICKEN_BITS_SCD_AUTO_RETRY_EN);

    /* Enable L1-Active */
    if (IwlTransContext->IwlConfig->Device_Family != IWL_DEVICE_FAMILY_8000)
        IwlClearBitsPrph(IwlTransContext, APMG_PCIDEV_STT_REG,
                         APMG_PCIDEV_STT_VAL_L1_ACT_DIS);

}

void
iwl_trans_pcie_tx_reset(
    struct iwl_trans_context *IwlTransContext
    )
{

    struct iwl_trans_pcie *trans_pcie = NULL;
    int txq_id;

    if(!IwlTransContext)
    {
        return;
    }
    trans_pcie = IWL_TRANS_GET_PCIE_TRANS(IwlTransContext);
    for (txq_id = 0; txq_id < IwlTransContext->IwlConfig->Base_Params->Num_Of_Queues;
         txq_id++) {
        struct iwl_txq *txq = &trans_pcie->txq[txq_id];

        iwl_write_direct32(IwlTransContext, FH_MEM_CBBC_QUEUE(txq_id),
                           (u32)(txq->q.dma_addr >> 8));
        iwl_pcie_txq_unmap(IwlTransContext, txq_id);
        txq->q.read_ptr = 0;
        txq->q.write_ptr = 0;
    }

    /* Tell NIC where to find the "keep warm" buffer */
    iwl_write_direct32(IwlTransContext, FH_KW_MEM_ADDR_REG,
                       (u32)(trans_pcie->kw.dma >> 4));

    iwl_pcie_tx_start(IwlTransContext, trans_pcie->Scd_Base_Addr);

}

/*
 * iwl_pcie_tx_stop - Stop all Tx DMA channels
 */
int
iwl_pcie_tx_stop(
    struct iwl_trans_context *IwlTransContext
    )
{

    struct iwl_trans_pcie *trans_pcie = IWL_TRANS_GET_PCIE_TRANS(IwlTransContext);
    int ch, txq_id;
    INTN ret;

    iwl_pcie_txq_set_sched(IwlTransContext, 0);

    /* Stop each Tx DMA channel, and wait for it to be idle */
    for (ch = 0; ch < FH_TCSR_CHNL_NUM; ch++) {
        iwl_write_direct32(IwlTransContext, FH_TCSR_CHNL_TX_CONFIG_REG(ch), 0x0);
        ret = IwlPollDirectBit(IwlTransContext,
                               FH_TSSR_TX_STATUS_REG,
                               FH_TSSR_TX_STATUS_REG_MSK_CHNL_IDLE(ch),
                               1000);
        if (ret < 0)

            DEBUG( (EFI_D_ERROR,"Failing on timeout while stopping DMA channel %d [0x%08x]\n",
                                                   ch,
                                                   IwlReadDirect32(IwlTransContext,
                                                                   FH_TSSR_TX_STATUS_REG)));
    }

    /*
     * This function can be called before the op_mode disabled the
     * queues. This happens when we have an rfkill interrupt.
     * Since we stop Tx altogether - mark the queues as stopped.
     */
    SetMem(trans_pcie->queue_stopped, sizeof(trans_pcie->queue_stopped),0);
    SetMem(trans_pcie->queue_used, sizeof(trans_pcie->queue_used),0);

    /* This can happen: start_hw, stop_device */
    if (!trans_pcie->txq)
        return 0;

    /* Unmap DMA from host system and free skb's */
    for (txq_id = 0; txq_id < IwlTransContext->IwlConfig->Base_Params->Num_Of_Queues;
         txq_id++)
        iwl_pcie_txq_unmap(IwlTransContext, txq_id);

    return 0;
}

/*
 * iwl_trans_tx_free - Free TXQ Context
 *
 * Destroy all TX DMA queues and structures
 */
void
iwl_pcie_tx_free(
    struct iwl_trans_context *IwlTransContext
    )
{
    int txq_id;
    struct iwl_trans_pcie *trans_pcie = NULL;

    if(!IwlTransContext)
    {
        return;
    }
    trans_pcie = IWL_TRANS_GET_PCIE_TRANS(IwlTransContext);
    /* Tx queues */
    if (trans_pcie->txq) {
        for (txq_id = 0;
             txq_id < IwlTransContext->IwlConfig->Base_Params->Num_Of_Queues; txq_id++)
            iwl_pcie_txq_free(IwlTransContext, txq_id);
    }
    if(trans_pcie->txq != NULL) {
        WifiFreePool(trans_pcie->txq,__FUNCTION__,__LINE__);
        trans_pcie->txq = NULL;
    }
    iwl_pcie_free_dma_ptr(IwlTransContext, &trans_pcie->kw);
    iwl_pcie_free_dma_ptr(IwlTransContext, &trans_pcie->scd_bc_tbls);
}

/*
 * iwl_pcie_tx_alloc - allocate TX context
 * Allocate all Tx DMA structures and initialize them
 */
static
int
iwl_pcie_tx_alloc(
    struct iwl_trans_context *IwlTransContext
    )
{

    int ret= 0;
    int txq_id, slots_num;
    EFI_STATUS Ret_Status;
    struct iwl_trans_pcie *trans_pcie = NULL;
    struct iwl_dma_ptr *scd_bd_tbls = NULL;
    struct iwl_dma_ptr *kw = NULL;
    u16 scd_bc_tbls_size = 0;
    INT32 retStatus = 0;


    if(!IwlTransContext)
    {
        return -1;
    }
    trans_pcie = IWL_TRANS_GET_PCIE_TRANS(IwlTransContext);
    scd_bd_tbls = &trans_pcie->scd_bc_tbls;
    kw = &trans_pcie->kw;
    scd_bc_tbls_size = (u16)(IwlTransContext->IwlConfig->Base_Params->Num_Of_Queues * sizeof(struct iwlagn_scd_bc_tbl));

    /*It is not allowed to alloc twice, so warn when this happens.
     * We cannot rely on the previous allocation, so free and fail */

    if ((trans_pcie->txq)) {
        DEBUG( (EFI_D_ERROR, "in alloc if loop\n" ));
        ret = -EINVAL;
        goto error;
    }

    //DMA Malloc Here

    Ret_Status = IwlTransContext->PciIoFuncs->AllocateBuffer(IwlTransContext->PciIoFuncs,
                                                             AllocateAnyPages,
                                                             EfiBootServicesData,
                                                             Check_num_of_pages(scd_bc_tbls_size),
                                                             (VOID **)&scd_bd_tbls->addr,
                                                             0);



    if (Ret_Status != 0) {
        DEBUG( (EFI_D_ERROR, "iwl_pcie_tx_alloc: Scheduler BC Table allocation failed,Ret_Status = %d\n",Ret_Status ));
        return -ENOMEM;
    }

    if (!scd_bd_tbls->addr)
        return -ENOMEM;

    retStatus = Dma_Map(IwlTransContext,
                         scd_bd_tbls->addr,
                         (UINT32) scd_bc_tbls_size,
                         TO_AND_FROM_DEVICE,
                         (UINT64*)(UINTN)scd_bd_tbls->addr,
                         &scd_bd_tbls->mapping);
    if (retStatus != 0) {
        DEBUG( (EFI_D_ERROR, "iwl_pcie_tx_alloc: Scheduler BC Table map failed with status = %x\n",retStatus ));
        goto error_scd_tbls;
    }



    Ret_Status = IwlTransContext->PciIoFuncs->AllocateBuffer(IwlTransContext->PciIoFuncs,
                                                             AllocateAnyPages,
                                                             EfiBootServicesData,
                                                             Check_num_of_pages(IWL_KW_SIZE),
                                                             (VOID **)&kw->addr,
                                                             0);



    if (Ret_Status != 0) {
        DEBUG( (EFI_D_ERROR, "iwl_pcie_tx_alloc: Keep Warm allocation failed, status = %d\n",Ret_Status ));
        goto error_scd_tbls;
    }

    retStatus = Dma_Map(IwlTransContext,
                         kw->addr,
                         (UINT32) IWL_KW_SIZE,
                         TO_AND_FROM_DEVICE,
                         (UINT64*)(UINTN)kw->addr,
                         &kw->mapping);
    if (retStatus != 0) {
        DEBUG( (EFI_D_ERROR, "iwl_pcie_tx_alloc: Keep Warm map failed with status = %x\n",retStatus ));
        goto error_kw_buf;
    }

    trans_pcie->txq = WifiAllocatepool(IwlTransContext->IwlConfig->Base_Params->Num_Of_Queues * (sizeof(struct iwl_txq)),__FUNCTION__);
    if(trans_pcie->txq == NULL) {
        DEBUG( (EFI_D_ERROR, "iwl_pcie_tx_alloc: Allocate pool failed\n" ));
        goto error;
    }

    /* Alloc and init all Tx queues, including the command queue (#4/#9) */
    for (txq_id = 0; txq_id < IwlTransContext->IwlConfig->Base_Params->Num_Of_Queues; txq_id++) {
        slots_num = (txq_id == trans_pcie->Cmd_Queue) ?
            TFD_CMD_SLOTS : TFD_TX_CMD_SLOTS;
        ret = iwl_pcie_txq_alloc(IwlTransContext, &trans_pcie->txq[txq_id],slots_num, txq_id);
        if (ret) {
            DEBUG( (EFI_D_ERROR, "iwl_pcie_tx_alloc:Tx %d queue alloc failed\n", txq_id ));
            goto error;
        }
    }

    return 0;

error_kw_buf:

     Ret_Status = IwlTransContext->PciIoFuncs->FreeBuffer(IwlTransContext->PciIoFuncs,
                                                           Check_num_of_pages(IWL_KW_SIZE),
                                                           kw->addr
                                                           );


    if (Ret_Status != 0) {
        DEBUG( (EFI_D_ERROR, "iwl_pcie_tx_alloc: kw->addr Memory release failed,status = %d\n",Ret_Status ));
    }
error_scd_tbls:

    Ret_Status = IwlTransContext->PciIoFuncs->FreeBuffer(IwlTransContext->PciIoFuncs,
                                                         Check_num_of_pages(scd_bc_tbls_size),
                                                         scd_bd_tbls->addr
                                                         );



    if (Ret_Status != 0) {
        DEBUG( (EFI_D_ERROR, "iwl_pcie_tx_alloc:scd_bd_tbls->addr Memory release failed,status = %d\n",Ret_Status ));
    }
    return (int)Ret_Status;

error:
    iwl_pcie_tx_free(IwlTransContext);
    return ret;
}

int
iwl_pcie_tx_init(
    struct iwl_trans_context *IwlTransContext
    )
{

    struct iwl_trans_pcie *trans_pcie = NULL;
    int ret;
    int txq_id, slots_num;
    bool alloc = false;

    if(!IwlTransContext)
    {
        return -1;
    }
    trans_pcie = IWL_TRANS_GET_PCIE_TRANS(IwlTransContext);
    IwlTransContext->IwlConfig->Base_Params->Num_Of_Queues=20;

    if (!trans_pcie->txq) {
        ret = iwl_pcie_tx_alloc(IwlTransContext);
        if (ret)
            goto error;
        alloc = true;
    }

    /* Turn off all Tx DMA fifos */
    IwlWritePrph(IwlTransContext, SCD_TXFACT, 0);

    /* Tell NIC where to find the "keep warm" buffer */

    iwl_write_direct32(IwlTransContext, FH_KW_MEM_ADDR_REG,
                       (u32)(trans_pcie->kw.dma >> 4));

    /* Alloc and init all Tx queues, including the command queue (#4/#9) */
    for (txq_id = 0; txq_id < IwlTransContext->IwlConfig->Base_Params->Num_Of_Queues;
         txq_id++) {
        slots_num = (txq_id == trans_pcie->Cmd_Queue) ?
            TFD_CMD_SLOTS : TFD_TX_CMD_SLOTS;
        ret = iwl_pcie_txq_init(IwlTransContext,
                                &trans_pcie->txq[txq_id],
                                slots_num,
                                txq_id);
        if (ret) {
            DEBUG( (EFI_D_ERROR, "iwl_pcie_tx_init:Tx %d queue init failed\n", txq_id ));
            goto error;
        }
    }

    return 0;
error:
    /*Upon error, free only if we allocated something */
    if (alloc)
        iwl_pcie_tx_free(IwlTransContext);
    return ret;
}

static
void
iwl_pcie_txq_progress(
    struct iwl_trans_pcie *trans_pcie,
    struct iwl_txq *txq
    )
{
    if(!trans_pcie || !txq)
    {
        return;
    }

    if (!trans_pcie->wd_timeout)
        return;
    /*
     * if empty delete timer, otherwise move timer forward
     * since we're making progress on this queue
     */
    if (txq->q.read_ptr == txq->q.write_ptr)
        ;

    else
        ;

}
/* This Api will be invoked when TX_CMD response is received and processed by the RX Interrupt
 * After the response is received the corresponding TX TFD entries will be freed
 * */
/* Frees buffers until index _not_ inclusive */
void
iwl_trans_pcie_reclaim(
    struct iwl_trans_context *IwlTransContext,
    int txq_id,
    int ssn,
    struct iwl_tx_skb *skb
    )
{

    struct iwl_trans_pcie *trans_pcie = NULL;
    struct iwl_txq *txq = NULL;
    /* n_bd is usually 256 => n_bd - 1 = 0xff */
    int tfd_num = 0;
    struct iwl_queue *q = NULL;
    int last_to_free;
    //	EFI_STATUS Status;

    if(!IwlTransContext || !skb)
    {
        return;
    }
    trans_pcie = IWL_TRANS_GET_PCIE_TRANS(IwlTransContext);
    txq = &trans_pcie->txq[txq_id];
    tfd_num = ssn & (txq->q.n_bd - 1);
    q = &txq->q;

    /* This function is not meant to release cmd queue*/
    if (WARN_ON(txq_id == trans_pcie->Cmd_Queue))
        return;

    if (!txq->active) {
        DEBUG( (EFI_D_INFO,"Q %d inactive - ignoring idx %d\n",
                                              txq_id, ssn));
        goto out;
    }

    if (txq->q.read_ptr == tfd_num)
        goto out;

    DEBUG( (EFI_D_INFO, "[Q %d] %d -> %d (%d ))\n",
                                           txq_id, txq->q.read_ptr, tfd_num, ssn));


    /*Since we free until index _not_ inclusive, the one before index is
     * the last we will free. This one must be used */
    last_to_free = iwl_queue_dec_wrap(tfd_num, q->n_bd);

    if (!iwl_queue_used(q, last_to_free)) {
        DEBUG( (EFI_D_INFO, "Read index for DMA queue txq id (%d )), last_to_free %d is out of range [0-%d] %d %d.\n",
                                               txq_id, last_to_free, q->n_bd,q->write_ptr, q->read_ptr));
        goto out;
    }

    for (;
         q->read_ptr != tfd_num;
         q->read_ptr = iwl_queue_inc_wrap(q->read_ptr, q->n_bd)) {


        CopyMem(skb->cb,txq->entries[txq->q.read_ptr].tx_skb.cb,48);

        CopyMem(&skb->header,&txq->entries[txq->q.read_ptr].tx_skb.header,sizeof(WIFI_Frame_Header));

        SetMem(&txq->entries[txq->q.read_ptr].tx_skb,sizeof(struct iwl_tx_skb),0);

        iwl_pcie_txq_inval_byte_cnt_tbl(IwlTransContext, txq);

        iwl_pcie_txq_free_tfd(IwlTransContext, txq);
    }

    iwl_pcie_txq_progress(trans_pcie, txq);

    if (iwl_queue_space(&txq->q) > txq->q.low_mark)
        ;
out:
    return;
}


/*
 * iwl_pcie_cmdq_reclaim - Reclaim TX command queue entries already Tx'd
 *
 * When FW advances 'R' index, all entries between old and new 'R' index
 * need to be reclaimed. As result, some free space forms.  If there is
 * enough free space (> low mark), wake the stack that feeds us.
 */
static
void
iwl_pcie_cmdq_reclaim(
    struct iwl_trans_context *IwlTransContext,
    int txq_id,
    int idx
    )
{

    struct iwl_trans_pcie *trans_pcie = NULL;
    struct iwl_txq *txq = NULL;
    struct iwl_queue *q = NULL;
    int nfreed = 0;

    if(!IwlTransContext)
    {
        return;
    }
    trans_pcie = IWL_TRANS_GET_PCIE_TRANS(IwlTransContext);
    txq = &trans_pcie->txq[txq_id];
    q = &txq->q;

    if ((idx >= q->n_bd) || (!iwl_queue_used(q, idx))) {
        DEBUG( (EFI_D_ERROR, "Read index for DMA queue txq id (%d )), index %d is out of range [0-%d] %d %d.\n",
                                                txq_id, idx, q->n_bd,
                                                q->write_ptr, q->read_ptr));

        return;
    }

    for (idx = iwl_queue_inc_wrap(idx, q->n_bd); q->read_ptr != idx;
         q->read_ptr = iwl_queue_inc_wrap(q->read_ptr, q->n_bd)) {

        if (nfreed++ > 0) {
            DEBUG( (EFI_D_INFO, "HCMD skipped: index (%d )) %d %d\n",
                                                   idx, q->write_ptr, q->read_ptr));

            IwlWritePrph(IwlTransContext, DEVICE_SET_NMI_REG, 1);
        }
    }

    if (q->read_ptr == q->write_ptr) {
        if(!trans_pcie->cmd_in_flight)
            DEBUG( (EFI_D_ERROR, "WARN ON for cmd_in_flight \r\n" ));

        trans_pcie->cmd_in_flight = false;

        Iwl_Clear_Bit(IwlTransContext,
                      CSR_GP_CNTRL,
                      CSR_GP_CNTRL_REG_FLAG_MAC_ACCESS_REQ);
    }

    iwl_pcie_txq_progress(trans_pcie, txq);
}

static
int
iwl_pcie_txq_set_ratid_map(
    struct iwl_trans_context *IwlTransContext,
    u16 ra_tid,
    u16 txq_id
    )
{
    struct iwl_trans_pcie *trans_pcie = IWL_TRANS_GET_PCIE_TRANS(IwlTransContext);
    u32 tbl_dw_addr;
    u32 tbl_dw;
    u16 scd_q2ratid;

    scd_q2ratid = ra_tid & SCD_QUEUE_RA_TID_MAP_RATID_MSK;

    tbl_dw_addr = trans_pcie->Scd_Base_Addr +
        SCD_TRANS_TBL_OFFSET_QUEUE(txq_id);

    tbl_dw = iwl_trans_read_mem32(IwlTransContext, tbl_dw_addr);

    if (txq_id & 0x1)
        tbl_dw = (scd_q2ratid << 16) | (tbl_dw & 0x0000FFFF);
    else
        tbl_dw = scd_q2ratid | (tbl_dw & 0xFFFF0000);

    iwl_trans_write_mem32(IwlTransContext, tbl_dw_addr, tbl_dw);

    return 0;
}

static
void
iwl_pcie_txq_set_inactive(
    struct iwl_trans_context *IwlTransContext,
    u16 txq_id
    )
{
    /* Simply stop the queue, but don't change any configuration;
     * the SCD_ACT_EN bit is the write-enable mask for the ACTIVE bit. */
    IwlWritePrph(IwlTransContext,
                 SCD_QUEUE_STATUS_BITS(txq_id),
                 (0 << SCD_QUEUE_STTS_REG_POS_ACTIVE)|
                 (1 << SCD_QUEUE_STTS_REG_POS_SCD_ACT_EN));
}

/* Receiver address (actually, Rx station's index into station table),
 * combined with Traffic ID (QOS priority), in format used by Tx Scheduler */
#define BUILD_RAxTID(sta_id, tid)   (((sta_id) << 4) + (tid))

VOID Iwl_Trans_Pcie_Txq_Enable (
    IN struct iwl_trans_context *IwlTransContext,
    IN UINT32 txq_id,
    IN UINT32 fifo,
    IN INTN sta_id,
    IN UINT32 tid,
    IN UINT32 frame_limit,
    IN UINT16 ssn
    )
{

    struct iwl_trans_pcie *trans_pcie = NULL;

    if(!IwlTransContext)
    {
        return;
    }
    trans_pcie = IWL_TRANS_GET_PCIE_TRANS(IwlTransContext);
    /*
    if (test_and_set_bit(txq_id, trans_pcie->queue_used))
        //WARN_ONCE(1, "queue %d already used - expect issues", txq_id);
        ErrorPrint(L"queue %d already used - expect issues", txq_id);
     */

    /* Stop this Tx queue before configuring it */
    iwl_pcie_txq_set_inactive(IwlTransContext, (UINT16)txq_id);

    /* Set this queue as a chain-building queue unless it is CMD queue */
    if (txq_id != trans_pcie->Cmd_Queue)
        IwlSetBitsPrph(IwlTransContext, SCD_QUEUECHAIN_SEL, BIT(txq_id));

    /* If this queue is mapped to a certain station: it is an AGG queue */
    if (sta_id >= 0) {
        UINT16 ra_tid = (UINT16)(BUILD_RAxTID(sta_id, tid));

        /* Map receiver-address / traffic-ID to this queue */
        iwl_pcie_txq_set_ratid_map(IwlTransContext, (UINT16)ra_tid, (UINT16)txq_id);

        /* enable aggregations for the queue */
        IwlSetBitsPrph(IwlTransContext, SCD_AGGR_SEL, BIT(txq_id));
        trans_pcie->txq[txq_id].ampdu = true;
    } else {
        /*
         * disable aggregations for the queue, this will also make the
         * ra_tid mapping configuration irrelevant since it is now a
         * non-AGG queue.
         */
        IwlClearBitsPrph(IwlTransContext, SCD_AGGR_SEL, BIT(txq_id));

        ssn = (UINT16)(trans_pcie->txq[txq_id].q.read_ptr);
    }

    /* Place first TFD at index corresponding to start sequence number.
     * Assumes that ssn_idx is valid (!= 0xFFF) */
    trans_pcie->txq[txq_id].q.read_ptr = (ssn & 0xff);
    trans_pcie->txq[txq_id].q.write_ptr = (ssn & 0xff);

    iwl_write_direct32(IwlTransContext, HBUS_TARG_WRPTR,
                       (ssn & 0xff) | (txq_id << 8));

    IwlWritePrph(IwlTransContext, SCD_QUEUE_RDPTR(txq_id), ssn);

    /* Set up Tx window size and frame limit for this queue */
    iwl_trans_write_mem32(IwlTransContext,
                          trans_pcie->Scd_Base_Addr + SCD_CONTEXT_QUEUE_OFFSET(txq_id),
                          0);
    iwl_trans_write_mem32(IwlTransContext, trans_pcie->Scd_Base_Addr +
                          SCD_CONTEXT_QUEUE_OFFSET(txq_id) + sizeof(u32),
                          ((frame_limit << SCD_QUEUE_CTX_REG2_WIN_SIZE_POS) &
                           SCD_QUEUE_CTX_REG2_WIN_SIZE_MSK) |
                          ((frame_limit << SCD_QUEUE_CTX_REG2_FRAME_LIMIT_POS) &
                           SCD_QUEUE_CTX_REG2_FRAME_LIMIT_MSK));

    /* Set up Status area in SRAM, map to Tx DMA/FIFO, activate the queue */
    IwlWritePrph(IwlTransContext, SCD_QUEUE_STATUS_BITS(txq_id),
                 (1 << SCD_QUEUE_STTS_REG_POS_ACTIVE) |
                 (fifo << SCD_QUEUE_STTS_REG_POS_TXF) |
                 (1 << SCD_QUEUE_STTS_REG_POS_WSL) |
                 SCD_QUEUE_STTS_REG_MSK);
    trans_pcie->txq[txq_id].active = true;

    DEBUG( (EFI_D_INFO,"Activate queue %d on FIFO %d WrPtr: %d\n",
                                          txq_id, fifo, ssn & 0xff));
}

void
iwl_trans_pcie_txq_disable(
    struct iwl_trans_context *IwlTransContext,
    int txq_id
    )
{

    struct iwl_trans_pcie *trans_pcie = NULL;
    u32 stts_addr = 0;
    static const u32 zero_val[4] = {0};

    if(!IwlTransContext)
    {
        return;
    }
    trans_pcie = IWL_TRANS_GET_PCIE_TRANS(IwlTransContext);
    stts_addr = trans_pcie->Scd_Base_Addr + SCD_TX_STTS_QUEUE_OFFSET(txq_id);

    /*
     * Upon HW Rfkill - we stop the device, and then stop the queues
     * in the op_mode. Just for the sake of the simplicity of the op_mode,
     * allow the op_mode to call txq_disable after it already called
     * stop_device.					`
     */

    if (!test_and_clear_bit(txq_id, trans_pcie->queue_used)) {

        if(test_bit(STATUS_DEVICE_ENABLED, &IwlTransContext->Status)) {

            DEBUG( (EFI_D_INFO, "queue %d not used \r\n",txq_id ));

        }
        return;
    }

    iwl_pcie_txq_set_inactive(IwlTransContext, (u16)txq_id);

    iwl_trans_write_mem(IwlTransContext, stts_addr, (void *)zero_val,
                        ARRAY_SIZEOF(zero_val));

    iwl_pcie_txq_unmap(IwlTransContext, txq_id);
    trans_pcie->txq[txq_id].ampdu = false;

    DEBUG( (EFI_D_INFO, "Deactivate queue %d\n",txq_id ));
}

/*************** HOST COMMAND QUEUE FUNCTIONS   *****/

/*
 * iwl_pcie_enqueue_hcmd - enqueue a uCode command
 * @priv: device private data point
 * @cmd: a pointer to the ucode command structure
 *
 * The function returns < 0 values to indicate the operation
 * failed. On success, it returns the index (>= 0) of command in the
 * command queue.
 */


/**
 * kmemdup - duplicate region of memory
 *
 * @src: memory region to duplicate
 * @len: memory region length
 * @gfp: GFP mask to use
 */

void
*duplicate_memory(
    const void *src,
    size_t len
    )
{
    void *p;
    //INT32 Status;

    p = WifiAllocatepool(len,__FUNCTION__);
    if(p == NULL) {
        DEBUG( (EFI_D_ERROR, "pcie:tx.c:error locating memory in duplicate memory API\n" ));
        return NULL;
    }
    //Add status checking code.

    CopyMem(p,src,len);

    return p;
}

static
int
iwl_pcie_enqueue_hcmd(
    struct iwl_trans_context *IwlTransContext,
    struct Iwl_Host_Cmd *cmd
    )
{
    struct iwl_trans_pcie *trans_pcie = NULL;
    struct iwl_txq *txq = NULL;
    struct iwl_queue *q = NULL;
    struct iwl_device_cmd *out_cmd;
    struct iwl_cmd_meta *out_meta;
    void *dup_buf = NULL;
    //dma_addr_t phys_addr;
    int idx = 0;
    u16 copy_size, cmd_size, scratch_size;
    bool had_nocopy = false;
    int i, ret;
    u32 cmd_pos;
    const u8 *cmddata[IWL_MAX_CMD_TBS_PER_TFD];
    u16 cmdlen[IWL_MAX_CMD_TBS_PER_TFD];
    EFI_STATUS status   = 0;
    INT32 retStatus = 0;
    int erridx   = 0;

    copy_size = sizeof(IWL_CMD_HEADER);
    cmd_size = sizeof(IWL_CMD_HEADER);

    if(!IwlTransContext || !cmd)
    {
        return -1;
    }
    trans_pcie = IWL_TRANS_GET_PCIE_TRANS(IwlTransContext);
    txq = &trans_pcie->txq[trans_pcie->Cmd_Queue];
    q = &txq->q;

    for (i = 0; i < IWL_MAX_CMD_TBS_PER_TFD; i++) {
        cmddata[i] = cmd->Data[i];
        cmdlen[i] = cmd->Length[i];

        if (!cmd->Length[i])
            continue;

        /* need at least IWL_HCMD_SCRATCHBUF_SIZE copied */
        if (copy_size < IWL_HCMD_SCRATCHBUF_SIZE) {
            int copy = IWL_HCMD_SCRATCHBUF_SIZE - copy_size;

            if (copy > cmdlen[i])
                copy = cmdlen[i];
            cmdlen[i] -= (u16)copy;
            cmddata[i] += copy;
            copy_size += (u16)copy;
        }
        //IWL_HCMD_DFL_NOCOPY is observed to be set for command packets
        if (cmd->DataFlags[i] & IWL_HCMD_DFL_NOCOPY) {
            had_nocopy = true;
            if ((cmd->DataFlags[i] & IWL_HCMD_DFL_DUP)) {
                DEBUG( (EFI_D_INFO, "WARN ON cmd->data_flags error\r\n" ));
                erridx = -EINVAL;
                goto free_dup_buf;
            }
        } else if (cmd->DataFlags[i] & IWL_HCMD_DFL_DUP) {
            /*
             * This is also a chunk that isn't copied
             * to the static buffer so set had_nocopy.
             */
            had_nocopy = true;

            /* only allowed once */
            if (WARN_ON(dup_buf)) {
                erridx = -EINVAL;
                goto free_dup_buf;
            }

            dup_buf = duplicate_memory(cmddata[i], cmdlen[i]);
            if (!dup_buf)
            {
                erridx = -ENOMEM;
                goto free_dup_buf;
            }
        } else {
            /* NOCOPY must not be followed by normal! */
            if (WARN_ON(had_nocopy)) {
                erridx = -EINVAL;
                goto free_dup_buf;
            }
            copy_size += cmdlen[i];
        }
        cmd_size += cmd->Length[i];
    }

    /*
     * If any of the command structures end up being larger than
     * the TFD_MAX_PAYLOAD_SIZE and they aren't dynamically
     * allocated into separate TFDs, then we will need to
     * increase the size of the buffers.
     */
    if ((copy_size > TFD_MAX_PAYLOAD_SIZE)) {

        DEBUG( (EFI_D_ERROR, "Command %s (%#x) is too large (%d bytes )) \r\n", \
                                                get_cmd_string(trans_pcie, cmd->Id), cmd->Id, copy_size));

        erridx = -EINVAL;
        goto free_dup_buf;
    }

    if (iwl_queue_space(q) < ((cmd->Flags & CMD_ASYNC) ? 2 : 1)) {
        DEBUG( (EFI_D_ERROR, "Trans--No space in command queue\n" ));

        iwl_op_mode_cmd_queue_full_Stub();
        erridx = -ENOSPC;
        goto free_dup_buf;
    }

    idx = get_cmd_index(q, q->write_ptr);

    SetMem(txq->entries[idx].cmd,sizeof(struct iwl_device_cmd),0);
    out_cmd = txq->entries[idx].cmd;


    out_meta = &txq->entries[idx].meta;
    SetMem(out_meta,sizeof(*out_meta),0);   /* re-initialize to NULL */

    for(i = 0; i < IWL_NUM_OF_TBS; i++)
    {
        txq->entries[idx].tb_cpuaddr[i] = NULL;
        txq->entries[idx].tb_phys[i] = 0;
    }


    if (cmd->Flags & CMD_WANT_SKB) {
        out_meta->source = cmd;
    }

    /* set up the header */

    out_cmd->Hdr.Cmd = cmd->Id;
    out_cmd->Hdr.Flags = 0;
    out_cmd->Hdr.Sequence =
        CONV_TO_LE16_BIT(QUEUE_TO_SEQ(trans_pcie->Cmd_Queue) |
                    INDEX_TO_SEQ(q->write_ptr));

    /* and copy the data that needs to be copied */
    //cmd_pos = offsetof_Iwl(struct iwl_device_cmd, Payload);
    cmd_pos = sizeof(IWL_CMD_HEADER);
    copy_size = sizeof(IWL_CMD_HEADER);


    for (i = 0; i < IWL_MAX_CMD_TBS_PER_TFD; i++) {
        int copy = 0;

        if (!cmd->Length[i])
            continue;

        /* need at least IWL_HCMD_SCRATCHBUF_SIZE copied */
        if (copy_size < IWL_HCMD_SCRATCHBUF_SIZE) {
            copy = IWL_HCMD_SCRATCHBUF_SIZE - copy_size;

            if (copy > cmd->Length[i])
                copy = cmd->Length[i];
        }

        /* copy everything if not nocopy/dup */
        if (!(cmd->DataFlags[i] & (IWL_HCMD_DFL_NOCOPY |
                                   IWL_HCMD_DFL_DUP))) {
            copy = cmd->Length[i];

            if (copy) {
                CopyMem((u8 *)out_cmd + cmd_pos, cmd->Data[i], copy);
                cmd_pos += copy;
                copy_size += (u16)copy;
                continue;
            }
        }
        /*
         * Otherwise we need at least IWL_HCMD_SCRATCHBUF_SIZE copied
         * in total (for the scratchbuf handling), but copy up to what
         * we can fit into the payload for debug dump purposes.
         */
        copy = min_t(int, TFD_MAX_PAYLOAD_SIZE - cmd_pos, cmd->Length[i]);

        CopyMem((u8 *)out_cmd + cmd_pos, cmd->Data[i], copy);
        cmd_pos += (UINT16)copy;

        /* However, treat copy_size the proper way, we need it below */
        if (copy_size < IWL_HCMD_SCRATCHBUF_SIZE) {
            copy = IWL_HCMD_SCRATCHBUF_SIZE - copy_size;

            if (copy > cmd->Length[i])
                copy = cmd->Length[i];
            copy_size += (UINT16)copy;
        }
    }

    DEBUG( (EFI_D_INFO, "Sending command (#%x )), seq: 0x%04X, %d bytes at %d[%d]:%d\n",
                                           out_cmd->Hdr.Cmd, le16_to_cpu(out_cmd->Hdr.Sequence),
                                           cmd_size, q->write_ptr, idx, trans_pcie->Cmd_Queue));

    /* start the TFD with the scratchbuf */
    scratch_size = min_t(int, copy_size, IWL_HCMD_SCRATCHBUF_SIZE);

    CopyMem(&txq->scratchbufs[idx], &out_cmd->Hdr, scratch_size);

    iwl_pcie_txq_build_tfd(IwlTransContext,
                           txq,
                           iwl_pcie_get_scratchbuf_dma(txq, idx),
                           scratch_size,
                           1);

    /* map first command fragment, if any remains */
    if (copy_size > scratch_size) {

        DEBUG( (EFI_D_INFO, "Before dma_map_single\n" ));
        //Convert virtual address to Physical address
        //and Map this Phy_addr and provide a pointer on the mapping area.

        status =  WirelessAllocatePage((VOID **)&txq->entries[idx].tb_cpuaddr[TB0]);

        if (EFI_ERROR(status))
        {
            DEBUG( (EFI_D_ERROR, "TX Enqueue cmd_addr AllocBuffer Failed,status = %d\n",status ));
            erridx = -ENOMEM;
            goto out;
        }
        CopyMem(txq->entries[idx].tb_cpuaddr[TB0], (UINT8 *)&out_cmd->Hdr + scratch_size, copy_size);

        txq->entries[idx].tb_phys[TB0] = dma_map_single(IwlTransContext,
                                                        trans_pcie->Cmd_Queue,
                                                        0,
                                                        (void *)txq->entries[idx].tb_cpuaddr[TB0],
                                                        PAGE_SIZE,
                                                        DMA_TO_DEVICE);

        if(!txq->entries[idx].tb_phys[TB0]) {
            DEBUG( (EFI_D_ERROR, "TX Enqueue dma_mapping_error\n" ));
            iwl_pcie_tfd_unmap_and_free(IwlTransContext, &txq->tfds[q->write_ptr], txq);
            erridx = -ENOMEM;
            goto free_dup_buf;
        }
        status = iwl_pcie_txq_build_tfd(IwlTransContext, txq, txq->entries[idx].tb_phys[TB0],
                                        copy_size - scratch_size, 0);
        if (status != 0)
        {
            iwl_pcie_tfd_unmap_and_free(IwlTransContext, &txq->tfds[q->write_ptr], txq);
            erridx = -ENOMEM;
            goto free_dup_buf;
        }
    }

    //phys_addr = 0;
    /* map the remaining (adjusted) nocopy/dup fragments */
    for (i = 0; i < IWL_MAX_CMD_TBS_PER_TFD; i++) {

        const void *data = cmddata[i];
        DEBUG( (EFI_D_INFO, "Before dma_map_single2 length %d\n", cmdlen[i] ));

        if (!cmdlen[i])
            continue;
        if (!(cmd->DataFlags[i] & (IWL_HCMD_DFL_NOCOPY |
                                   IWL_HCMD_DFL_DUP)))
            continue;
        if (cmd->DataFlags[i] & IWL_HCMD_DFL_DUP)
            data = dup_buf;


        status =  WirelessAllocatePage((VOID **)&txq->entries[idx].tb_cpuaddr[i+1]);

        if (EFI_ERROR(status))
        {
            DEBUG( (EFI_D_ERROR, "TX Enqueue tb1_cpuaddr AllocBuffer Failed, status = %d,i : %d\n",status,i ));
            erridx = -ENOMEM;
            goto out;
        }
        CopyMem(txq->entries[idx].tb_cpuaddr[i+1], data, cmdlen[i]);

        retStatus = Dma_Map(IwlTransContext,
                         (txq->entries[idx].tb_cpuaddr[i+1]),
                         (UINT32) cmdlen[i],
                         DMA_TO_DEVICE,
                         (UINT64*)(UINTN) &txq->entries[idx].tb_phys[i+1],
                         &txq->tbs_mapping[idx][i+1]);

        if (retStatus != 0) {
            DEBUG( (EFI_D_ERROR, "iwl_pcie_enqueue_hcmd: DMA Map Failed for tb1_cpuaddr,status = %d\n",retStatus ));
            iwl_pcie_tfd_unmap_and_free(IwlTransContext, &txq->tfds[q->write_ptr], txq);
            erridx = -ENOMEM;
            goto out;
        }
        if(!txq->entries[idx].tb_phys[i+1]) {
            iwl_pcie_tfd_unmap_and_free(IwlTransContext, &txq->tfds[q->write_ptr], txq);
            erridx = -ENOMEM;
            goto out;
        }
        status = iwl_pcie_txq_build_tfd(IwlTransContext, txq, txq->entries[idx].tb_phys[i+1], cmdlen[i], 0);
        if (status != 0)
        {
            iwl_pcie_tfd_unmap_and_free(IwlTransContext, &txq->tfds[q->write_ptr], txq);
            erridx = -ENOMEM;
            goto out;
        }
        //phys_addr = 0;
    }

    out_meta->flags = (u32)(cmd->Flags);
    if (WARN_ON_ONCE(txq->entries[idx].free_buf))
    {
        DEBUG( (EFI_D_ERROR, " txq->entries[idx].free_buf = dup_buf \n " ));
        if (dup_buf != NULL)
        {
            WifiFreePool((VOID *)txq->entries[idx].free_buf,__FUNCTION__,__LINE__);
            txq->entries[idx].free_buf = dup_buf;
        }
    }

    txq->need_update = 1;

    /*
     * wake up the NIC to make sure that the firmware will see the host
     * command - we will let the NIC sleep once all the host commands
     * returned.
     */
    if (!trans_pcie->cmd_in_flight) {
        trans_pcie->cmd_in_flight = true;

        Iwl_Trans_Pcie_Set_Bits_Mask(IwlTransContext,
                                      CSR_GP_CNTRL,
                                      CSR_GP_CNTRL_REG_FLAG_MAC_ACCESS_REQ,
                                      CSR_GP_CNTRL_REG_FLAG_MAC_ACCESS_REQ);

        ret = Iwl_Poll_Bit(IwlTransContext,
                           CSR_GP_CNTRL,
                           CSR_GP_CNTRL_REG_VAL_MAC_ACCESS_EN,
                           (CSR_GP_CNTRL_REG_FLAG_MAC_CLOCK_READY |
                            CSR_GP_CNTRL_REG_FLAG_GOING_TO_SLEEP),
                           15000);

        if (ret < 0) {
            Iwl_Clear_Bit(IwlTransContext, CSR_GP_CNTRL,
                          CSR_GP_CNTRL_REG_FLAG_MAC_ACCESS_REQ);
            trans_pcie->cmd_in_flight = false;
            erridx = -EIO;
            goto out;
        }
    }

    /* Increment and update queue's write index */
    q->write_ptr = iwl_queue_inc_wrap(q->write_ptr, q->n_bd);
    iwl_pcie_txq_inc_wr_ptr(IwlTransContext, txq);
    goto out;

out:
free_dup_buf:
    if ((erridx < 0) && (dup_buf != NULL))
    {
        if (dup_buf == txq->entries[idx].free_buf)
        {
            txq->entries[idx].free_buf = NULL;
        }
        WifiFreePool(dup_buf,__FUNCTION__,__LINE__);
        dup_buf = NULL;
    }
    return erridx < 0 ? erridx : idx;
}

/*
 * iwl_pcie_hcmd_complete - Pull unused buffers off the queue and reclaim them
 * @rxb: Rx buffer to reclaim
 * @handler_status: return value of the handler of the command
 *	(put in setup_rx_handlers)
 *
 * If an Rx buffer has an async callback associated with it the callback
 * will be executed.  The attached skb (if present) will only be freed
 * if the callback returns 1
 */



void
iwl_pcie_hcmd_complete(
    struct iwl_trans_context *IwlTransContext,
    struct iwl_rx_cmd_buffer *rxb,
    int handler_status
    )
{

    struct Iwl_Rx_Packet *pkt = NULL;
    u16 sequence = 0;
    int txq_id = 0;
    int index = 0;
    int cmd_index;
    struct iwl_device_cmd *cmd;
    struct iwl_cmd_meta *meta;
    struct iwl_trans_pcie *trans_pcie = NULL;
    struct iwl_txq *txq = NULL;

    if(!IwlTransContext || !rxb)
    {
        return;
    }

    pkt = (struct Iwl_Rx_Packet *)((UINT8 *)rxb->cpu_addr + rxb->_offset);
    sequence = le16_to_cpu(pkt->Hdr.Sequence);
    txq_id = SEQ_TO_QUEUE(sequence);
    index = SEQ_TO_INDEX(sequence);
    trans_pcie = IWL_TRANS_GET_PCIE_TRANS(IwlTransContext);
    txq = &trans_pcie->txq[trans_pcie->Cmd_Queue];

    /* If a Tx command is being handled and it isn't in the actual
     * command queue then there a command routing bug has been introduced
     * in the queue management code. */

    if ((txq_id != trans_pcie->Cmd_Queue)) {

        DEBUG( (EFI_D_INFO, "wrong command queue %d (should be %d )), sequence 0x%X readp=%d writep=%d\n", \
                                               txq_id, trans_pcie->Cmd_Queue, sequence,
                                               trans_pcie->txq[trans_pcie->Cmd_Queue].q.read_ptr,
                                               trans_pcie->txq[trans_pcie->Cmd_Queue].q.write_ptr));
        return;
    }

    cmd_index = get_cmd_index(&txq->q, index);
    cmd = txq->entries[cmd_index].cmd;
    meta = &txq->entries[cmd_index].meta;

    iwl_pcie_tfd_unmap_and_free(IwlTransContext, &txq->tfds[index], txq);

    /* Input error checking is done when commands are added to queue. */

    /* If WANT_SKB flag is set then the response needs to be held for the sake of
     * validating the response.
     * */
    if (meta->flags & CMD_WANT_SKB) {

        if(meta->source->Resp_Pkt != NULL)
            CopyMem(meta->source->Resp_Pkt,pkt,MAX_TXRESP_PKT_SIZE);
        meta->source->_rx_page_order = trans_pcie->rx_page_order;
        meta->source->Handler_Status = handler_status;
    }

    iwl_pcie_cmdq_reclaim(IwlTransContext, txq_id, index);

    if (!(meta->flags & CMD_ASYNC)) {
        if (!test_bit(STATUS_SYNC_HCMD_ACTIVE, &IwlTransContext->Status)) {
            DEBUG( (EFI_D_INFO, "IwlTransContext--HCMD_ACTIVE already clear for command %x\n",cmd->Hdr.Cmd ));
        }

        clear_bit(STATUS_SYNC_HCMD_ACTIVE, &IwlTransContext->Status);


    }

    meta->flags = 0;
}

#define HOST_COMPLETE_TIMEOUT   (2 * HZ)

static
int
iwl_pcie_send_hcmd_async(
    struct iwl_trans_context *IwlTransContext,
    struct Iwl_Host_Cmd *cmd
    )
{

    struct iwl_trans_pcie *trans_pcie = NULL;
    UINT8 *temp = NULL;
    int ret;

    if(!IwlTransContext || !cmd)
    {
        return -1;
    }
    trans_pcie = IWL_TRANS_GET_PCIE_TRANS(IwlTransContext);
    temp = (UINT8*)trans_pcie;
    /* An asynchronous command can not expect an SKB to be set. */
    if (WARN_ON(cmd->Flags & CMD_WANT_SKB))
        return -EINVAL;

    ret = iwl_pcie_enqueue_hcmd(IwlTransContext, cmd);
    if (ret < 0) {
        DEBUG( (EFI_D_INFO,"Trans -- Error sending %s: enqueue_hcmd failed: %d\n",
                                              get_cmd_string(trans_pcie, cmd->Id), ret));
        return ret;
    }

    return 0;
}


static
void
iwl_trans_fw_error(
    struct iwl_trans_context *IwlTransContext
    )
{
    if(!IwlTransContext)
      return;

    /* prevent double restarts due to the same erroneous FW */
    if (!test_and_set_bit(STATUS_FW_ERROR, &IwlTransContext->Status))
        Iwl_Mvm_Nic_Error_Stub();
}

static
INTN
iwl_pcie_send_hcmd_sync(
    struct iwl_trans_context *IwlTransContext,
    struct Iwl_Host_Cmd *cmd
    )
{

    struct iwl_trans_pcie *trans_pcie = NULL;
    int cmd_idx;
    INTN ret = 0;
    int i=0;

    if(!IwlTransContext || !cmd)
    {
        return -1;
    }

    trans_pcie = IWL_TRANS_GET_PCIE_TRANS(IwlTransContext);

    if ((test_and_set_bit(STATUS_SYNC_HCMD_ACTIVE,
                          &IwlTransContext->Status))) {

        DEBUG( (EFI_D_ERROR, "Command %s: a command is already active!\n", get_cmd_string(trans_pcie, cmd->Id) ));

        return -EIO;
    }

    DEBUG( (EFI_D_INFO, "Trans--Setting HCMD_ACTIVE for command %x\n",cmd->Id ));

    cmd_idx = iwl_pcie_enqueue_hcmd(IwlTransContext, cmd);
    if (cmd_idx < 0) {
        ret = cmd_idx;
        clear_bit(STATUS_SYNC_HCMD_ACTIVE, &IwlTransContext->Status);
        DEBUG( (EFI_D_ERROR, "Error sending %x: enqueue_hcmd failed: %d\n",cmd->Id, ret ));
        return ret;
    }

    //Wait here for 200ms for the response for the command
    do {

        gBS->Stall(10000);
        i++;
        if (!test_bit(STATUS_SYNC_HCMD_ACTIVE,&IwlTransContext->Status))
            break;


    } while (i<=200);

    if (test_bit(STATUS_SYNC_HCMD_ACTIVE,&IwlTransContext->Status))
    {
        struct iwl_txq *txq = &trans_pcie->txq[trans_pcie->Cmd_Queue];
        struct iwl_queue *q = &txq->q;
        UINT8   *temp = NULL;

        temp =(UINT8*)q;

        DEBUG( (EFI_D_ERROR, "Error sending Command %x: time out at NIC\n",cmd->Id ));
        DEBUG( (EFI_D_INFO,"Current CMD queue read_ptr %d write_ptr %d\n",
                                              q->read_ptr, q->write_ptr));

        clear_bit(STATUS_SYNC_HCMD_ACTIVE, &IwlTransContext->Status);


        ret = -EDEVERROR;

        //IwlWritePrph(trans, DEVICE_SET_NMI_REG, 1);
        Iwl_Force_Nmi(IwlTransContext);
        iwl_trans_fw_error(IwlTransContext);

        goto cancel;
    }

    if (test_bit(STATUS_FW_ERROR, &IwlTransContext->Status)) {

        DEBUG( (EFI_D_ERROR,"Trans-- FW error in SYNC CMD %s\n",
                                               get_cmd_string(trans_pcie, cmd->Id)));

        ret = -EIO;
        goto cancel;
    }

    if (!(cmd->Flags & CMD_SEND_IN_RFKILL) &&
        test_bit(STATUS_RFKILL, &IwlTransContext->Status)) {

        DEBUG( (EFI_D_ERROR, "IwlTransContext-- RFKILL in SYNC CMD... no rsp\n" ));

        ret = -ERFKILL;
        goto cancel;
    }

    if ((cmd->Flags & CMD_WANT_SKB) && !cmd->Resp_Pkt) {
        DEBUG( (EFI_D_ERROR,"IwlTransContext -- Error: Response NULL in '%s'\n",
                                               get_cmd_string(trans_pcie, cmd->Id)));

        ret = -EIO;
        goto cancel;
    }

    return 0;

cancel:
    if (cmd->Flags & CMD_WANT_SKB) {
        /*
         * Cancel the CMD_WANT_SKB flag for the cmd in the
         * TX cmd queue. Otherwise in case the cmd comes
         * in later, it will possibly set an invalid
         * address (cmd->meta.source).
         */
        trans_pcie->txq[trans_pcie->Cmd_Queue].
        entries[cmd_idx].meta.flags &= ~CMD_WANT_SKB;
    }

    if (cmd->Resp_Pkt) {
        cmd->Resp_Pkt = NULL;
    }

    return ret;
}

INTN
iwl_trans_pcie_send_hcmd(
    struct iwl_trans_context *IwlTransContext,
    struct Iwl_Host_Cmd *cmd
    )
{
    if(!IwlTransContext || !cmd)
    {
        return -1;
    }

    if (!(cmd->Flags & CMD_SEND_IN_RFKILL) &&
        test_bit(STATUS_RFKILL, &IwlTransContext->Status)) {

        DEBUG( (EFI_D_INFO, "Dropping CMD 0x%x: RF KILL\r\n", cmd->Id ));
        return -ERFKILL;
    }

    if (cmd->Flags & CMD_ASYNC)
        return iwl_pcie_send_hcmd_async(IwlTransContext, cmd);

    /* We still can fail on RFKILL that can be asserted while we wait */
    return iwl_pcie_send_hcmd_sync(IwlTransContext, cmd);
}


/*This api is mainly responsible for enqueuing a data packet into tx hardware queue for transmission
 * to AP.
 * Other than command queue all the remaining tx queues have 256 txq entries,256 scratchbufs anf 256 tfd's
 * In the current implementation only 3 tfd's are used.
 * tb0-for holding scratch buffer,
 * tb1-excess header in the tx command other than the scratch buffer,
 * tb2-holds the actual payload.
 * After building the tfd appropriately the write ptr is incremented so that the NIC will be
 * notified about the new packet to be transmitted
 * */

INTN
iwl_trans_pcie_tx(
    struct iwl_trans_context *IwlTransContext,
    SK_BUFF *skb,
    struct iwl_device_cmd *dev_cmd,
    INTN txq_id
    )
{
    struct iwl_trans_pcie *trans_pcie = NULL;
    WIFI_Frame_Header *hdr = NULL;
    struct iwl_tx_cmd *tx_cmd = NULL;
    __le16 fc = 0;
    u8 hdr_len = 0;
    struct iwl_cmd_meta *out_meta;
    struct iwl_txq *txq;
    struct iwl_queue *q;
    dma_addr_t scratch_phys;
    void *tbdata_addr[3];
    u16 len =0, tb_len[3] = {0,0,0};
    u16 wifi_seq;
    int i;
    EFI_STATUS Status;

    if(!IwlTransContext || !skb || !dev_cmd)
    {
        return -1;
    }

    trans_pcie = IWL_TRANS_GET_PCIE_TRANS(IwlTransContext);
    hdr = (WIFI_Frame_Header *)skb->Data;
    tx_cmd = (struct iwl_tx_cmd *)dev_cmd->Payload;
    fc = hdr->FrameControl;
    hdr_len = (u8)GetHeaderLength(fc);


    txq = &trans_pcie->txq[txq_id];
    q = &txq->q;

    //If no space available in txq discard that packet and inform SNP
    //that packet transmission is failed.
    if (iwl_queue_space(q) < q->high_mark) {
        // iwl_stop_queue(IwlTransContext, txq);
        goto out_err;
    }

    wifi_seq = (u16)(WIFI_DRV_SEQUENCE_TO_SN(le16_to_cpu(hdr->Ieee80211Seq)));

    if((txq->ampdu && (wifi_seq & 0xff) != q->write_ptr))
        DEBUG( (EFI_D_INFO, "Q: %d WiFi Seq %d tfdNum %d \r\n",  txq_id, wifi_seq, q->write_ptr ));

    /* Set up driver data for this TFD */
    CopyMem(txq->entries[q->write_ptr].tx_skb.cb,skb->CtrlBuff,sizeof(skb->CtrlBuff));

    CopyMem(&txq->entries[q->write_ptr].tx_skb.header,(WIFI_Frame_Header *)skb->Data,sizeof(WIFI_Frame_Header));

    txq->entries[q->write_ptr].cmd = dev_cmd;

    for(i = 0; i < IWL_NUM_OF_TBS; i++)
    {
        txq->entries[q->write_ptr].tb_cpuaddr[i] = NULL;
        txq->entries[q->write_ptr].tb_phys[i] = 0;
    }

    dev_cmd->Hdr.Sequence = CONV_TO_LE16_BIT((u16)(QUEUE_TO_SEQ(txq_id) | INDEX_TO_SEQ(q->write_ptr)));

    /* the first tb holds the scratch buff. As per linux implementation, it should
     * neither be freed nor unmapped until stopping the NIC
     */

    txq->entries[q->write_ptr].tb_phys[TB0] = iwl_pcie_get_scratchbuf_dma(txq, q->write_ptr);

    scratch_phys = (dma_addr_t)(txq->entries[q->write_ptr].tb_phys[TB0]
                                + sizeof(struct Iwl_Cmd_Header)
                                + offsetof_Iwl(struct iwl_tx_cmd, scratch));

    tx_cmd->dram_lsb_ptr = (u32)(CONV_TO_LE32_BIT(scratch_phys));
    tx_cmd->dram_msb_ptr = iwl_get_dma_hi_addr(scratch_phys);

    /* out_meta will  */
    out_meta = &txq->entries[q->write_ptr].meta;

    /*
     * The second TB (tb1) points to the remainder of the TX command
     * and the 802.11 header - dword aligned size
     * (This calculation modifies the TX command, so do it before the
     * setup of the first TB)
     */
    len = sizeof(struct iwl_tx_cmd) + sizeof(struct Iwl_Cmd_Header)
        - IWL_HCMD_SCRATCHBUF_SIZE;

    tb_len[TB1] = ALIGN(len, 4);

    /* Tell NIC about any 2-byte padding after MAC header */
    if (tb_len[TB1] != len)
        tx_cmd->tx_flags |= TX_CMD_FLG_MH_PAD_MSK;

    /* The first TB points to the scratchbuf data - min_copy bytes */

    CopyMem(&txq->scratchbufs[q->write_ptr], &dev_cmd->Hdr,
            IWL_HCMD_SCRATCHBUF_SIZE);

    iwl_pcie_txq_build_tfd(IwlTransContext, txq, txq->entries[q->write_ptr].tb_phys[TB0],
                           IWL_HCMD_SCRATCHBUF_SIZE, true);

    /* map the data for TB1 */
    tbdata_addr[TB1] = ((u8 *)&dev_cmd->Hdr) + IWL_HCMD_SCRATCHBUF_SIZE;

    tb_len[TB2] = (u16)(skb->Len - hdr_len);

    tbdata_addr[TB2] = /*(u8 *)*/ skb->Data + hdr_len;

    /* For now only 2 tbs other than scratch buf are supported in data path*/

    for(i = 1; i <= IWL_MAX_CMD_TBS_PER_TFD; i++)
    {
        if(tb_len[i] != 0)
        {

            Status =  WirelessAllocatePage((VOID **)&txq->entries[q->write_ptr].tb_cpuaddr[i]);

            if(EFI_ERROR(Status)) {
                DEBUG( (EFI_D_ERROR, "Error in allocating PCI DMA memory for tb_cpuaddr TX_CMD,status : %d addr : %p\n",Status,txq->entries[q->write_ptr].tb_cpuaddr[i] ));
                return -1;
            }

            CopyMem(txq->entries[q->write_ptr].tb_cpuaddr[i], tbdata_addr[i], tb_len[i]);

            txq->entries[q->write_ptr].tb_phys[i] = dma_map_single(IwlTransContext, txq_id, (UINT8)i, txq->entries[q->write_ptr].tb_cpuaddr[i], tb_len[i], DMA_TO_DEVICE);

            if (!txq->entries[q->write_ptr].tb_phys[i])
            {
                DEBUG( (EFI_D_ERROR, "out_err :: 1\n" ));
                goto out_err;
            }
            Status = iwl_pcie_txq_build_tfd(IwlTransContext, txq,txq->entries[q->write_ptr].tb_phys[i],tb_len[i], false);
            if (Status != 0)
            {
                DEBUG( (EFI_D_ERROR, "out_err :: iwl_pcie_txq_build_tfd failed\n" ));
                iwl_pcie_tfd_unmap_and_free(IwlTransContext, &txq->tfds[q->write_ptr], txq);
                goto out_err;
            }
        }
    }

    /* Set up entry for this TFD in Tx byte-count array */
    iwl_pcie_txq_update_byte_cnt_tbl(IwlTransContext, txq, le16_to_cpu(tx_cmd->len));


    if (!CheckMoreFragsFlag(fc)) {
        txq->need_update = 1;
    } else {
        txq->need_update = 0;
    }

    /* start timer if queue currently empty */

    if (txq->need_update && q->read_ptr == q->write_ptr &&
        trans_pcie->wd_timeout)
        ;

    /* Tell device the write index *just past* this latest filled TFD */
    q->write_ptr = iwl_queue_inc_wrap(q->write_ptr, q->n_bd);
    iwl_pcie_txq_inc_wr_ptr(IwlTransContext, txq);

    /*
     * At this point the frame is "transmitted" successfully
     * and we will get a TX status notification eventually,
     * regardless of the value of ret. "ret" only indicates
     * whether or not we should update the write pointer.
     */

    return 0;

out_err:
    return -1;
}
