/**
 * @file iwl-trans-context.h
 *
 * @details
 * This header file contains function prototypes, structures related
 * to transmit bits through Intel NIC, etc. for Intel UNDI driver.
 *
 * @copyright
 * Copyright (c) 2017 Dell Inc. All rights reserved. This
 * software and associated documentation (if any) is furnished
 * under a license and may only be used or copied in accordance
 * with the terms of the license. Except as permitted by such
 * license, no part of this software or documentation may be
 * reproduced, stored in a retrieval system, or transmitted in
 * any form or by any means without the express written consent
 * of Dell Inc.
 *
 * Dell elects to use only the BSD license for any software where
 * a choice of BSD/GPLv2 license versions are made available with
 * the language indicating that BSD/GPLv2 or any later version may
 * be used, or where a choice of which version of the BSD/GPLv2 is
 * applied is unspecified.
 */

/******************************************************************************
 *
 * This file is provided under a dual BSD/GPLv2 license.  When using or
 * redistributing this file, you may do so under either license.
 *
 * GPL LICENSE SUMMARY
 *
 * Copyright(c) 2007 - 2014 Intel Corporation. All rights reserved.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of version 2 of the GNU General Public License as
 * published by the Free Software Foundation.
 *
 * This program is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110,
 * USA
 *
 * The full GNU General Public License is included in this distribution
 * in the file called COPYING.
 *
 * Contact Information:
 *  Intel Linux Wireless <ilw@linux.intel.com>
 * Intel Corporation, 5200 N.E. Elam Young Parkway, Hillsboro, OR 97124-6497
 *
 * BSD LICENSE
 *
 * Copyright(c) 2005 - 2014 Intel Corporation. All rights reserved.
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 *
 *  * Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 *  * Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in
 *    the documentation and/or other materials provided with the
 *    distribution.
 *  * Neither the name Intel Corporation nor the names of its
 *    contributors may be used to endorse or promote products derived
 *    from this software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
 * "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
 * LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
 * A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
 * OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 * SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
 * LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
 * DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
 * THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
 * OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 *
 *****************************************************************************/

#ifndef IWLTRANSCONTEXT_H_
#define IWLTRANSCONTEXT_H_

#include "iwl-config.h"
#include "iwl-fw.h"
#include "IwlErrno.h"
#include "iwl-op-mode.h"
#include <Protocol/PciIo.h>
#include <Softmac80211Lib.h>
#include <TransType.h>


/******************************************************************/
/***
*                           MACROS
*/
/******************************************************************/
#define DEF_CMD_PAYLOAD_SIZE 4096
#define MAX_NO_RECLAIM_CMDS 6

/**
 * DOC: Host command section
 *
 * A host command is a commaned issued by the upper layer to the fw. There are
 * several versions of fw that have several APIs. The transport layer is
 * completely agnostic to these differences.
 * The transport does provide helper functionnality (i.e. SYNC / ASYNC mode),
 */
#define SEQ_TO_QUEUE(s) (((s) >> 8) & 0x1f)
#define QUEUE_TO_SEQ(q) (((q) & 0x1f) << 8)
#define SEQ_TO_INDEX(s) ((s) & 0xff)
#define INDEX_TO_SEQ(i) ((i) & 0xff)
#define SEQ_RX_FRAME    CONV_TO_LE16_BIT(0x8000)

/*
 * number of transfer buffers (fragments) per transmit frame descriptor;
 * this is just the driver's idea, the hardware supports 20
 */
#define IWL_MAX_CMD_TBS_PER_TFD 2


#define IWL_MASK(lo, hi) ((1 << (hi)) | ((1 << (hi)) - (1 << (lo))))

/**
 * enum iwl_trans_status: transport status flags
 * @STATUS_SYNC_HCMD_ACTIVE: a SYNC command is being processed
 * @STATUS_DEVICE_ENABLED: APM is enabled
 * @STATUS_TPOWER_PMI: the device might be asleep (need to wake it up)
 * @STATUS_INT_ENABLED: interrupts are enabled
 * @STATUS_RFKILL: the HW RFkill switch is in KILL position
 * @STATUS_FW_ERROR: the fw is in error state
 * @STATUS_TRANS_GOING_IDLE: shutting down the trans, only special commands
 *  are sent
 * @STATUS_TRANS_IDLE: the trans is idle - general commands are not to be sent
 */
enum iwl_trans_status {
    STATUS_SYNC_HCMD_ACTIVE,
    STATUS_DEVICE_ENABLED,
    STATUS_TPOWER_PMI,
    STATUS_INT_ENABLED,
    STATUS_RFKILL,
    STATUS_FW_ERROR,
    STATUS_TRANS_GOING_IDLE,
    STATUS_TRANS_IDLE,
};

/*
 * Maximum number of HW queues the transport layer
 * currently supports
 */
#define IWL_MAX_HW_QUEUES   32
#define IWL_MAX_TID_COUNT   8
#define IWL_FRAME_LIMIT     64

/* iwl_cmd_header flags value */
#define IWL_CMD_FAILED_MSK 0x40

#define DEFAULT                             0
#define PROCESSED_AND_TO_BE_FREED           1
#define ENQUEUED_AND_IN_USE                 2

#define FH_RSCSR_FRAME_SIZE_MSK     0x00003FFF  /* bits 0-13 */
#define FH_RSCSR_FRAME_INVALID      0x55550000
#define FH_RSCSR_FRAME_ALIGN        0x40
#define IWL_MAX_TID_COUNT   8
/******************************************************************/
/***
*                           Typedefs
*/
/******************************************************************/

typedef struct iwl_trans_pcie IWL_TRANS_PCIE;
typedef struct iwl_cfg IWL_CONFIG;
typedef struct Iwl_Host_Cmd IWL_HOST_CMD;
typedef struct _SK_BUFF SK_BUFF;
typedef UINTN size_t;

typedef struct iwl_trans_context
{
    IWL_TRANS_PCIE *TransPcie;
    EFI_PCI_IO_PROTOCOL *PciIoFuncs;
    IWL_CONFIG *IwlConfig;
    unsigned long Status;
}IWL_TRANS_CONTEXT;


struct iwl_rx_cmd_buffer {
    UINT64 *cpu_addr;
    INTN _offset;
    BOOLEAN _page_stolen;
    UINT32 _rx_page_order;
    UINTN truesize;
};

typedef struct iwl_tx_skb {
    char cb[48];
    WIFI_Frame_Header header;
}IWL_TX_SKB;

/**
 * struct iwl_cmd_header
 *
 * This header format appears in the beginning of each command sent from the
 * driver, and each response/notification received from uCode.
 */
typedef struct Iwl_Cmd_Header {
    UINT8 Cmd;      /* Command ID:  REPLY_RXON, etc. */
    UINT8 Flags;    /* 0:5 reserved, 6 abort, 7 internal */
    /*
     * The driver sets up the sequence number to values of its choosing.
     * uCode does not use this value, but passes it back to the driver
     * when sending the response to each driver-originated command, so
     * the driver can match the response to the command.  Since the values
     * don't get used by uCode, the driver may set up an arbitrary format.
     *
     * There is one exception:  uCode sets bit 15 when it originates
     * the response/notification, i.e. when the response/notification
     * is not a direct response to a command sent by the driver.  For
     * example, uCode issues REPLY_RX when it sends a received frame
     * to the driver; it is not a direct response to any driver command.
     *
     * The Linux driver uses the following format:
     *
     *  0:7     tfd index - position within TX queue
     *  8:12    TX queue id
     *  13:14   reserved
     *  15      unsolicited RX or uCode-originated notification
     */
    UINT16 Sequence;
}IWL_CMD_HEADER;


typedef struct iwl_device_cmd {
    IWL_CMD_HEADER Hdr; /* uCode API */
    UINT8 Payload[DEF_CMD_PAYLOAD_SIZE];
}IWL_DEVICE_CMD;

#define TFD_MAX_PAYLOAD_SIZE (sizeof(struct iwl_device_cmd))

typedef struct Iwl_Rx_Packet {
    /*
     * The first 4 bytes of the RX frame header contain both the RX frame
     * size and some flags.
     * Bit fields:
     * 31:    flag flush RB request
     * 30:    flag ignore TC (terminal counter) request
     * 29:    flag fast IRQ request
     * 28-14: Reserved
     * 13-00: RX frame size
     */
    UINT32 len_n_flags;
    IWL_CMD_HEADER Hdr;
    UINT8 Data[];
}IWL_RX_PACKET;

/**
 * enum iwl_trans_state - state of the transport layer
 *
 * @IWL_TRANS_NO_FW: no fw has sent an alive response
 * @IWL_TRANS_FW_ALIVE: a fw has sent an alive response
 */
typedef enum iwl_trans_state {
    IWL_TRANS_NO_FW = 0,
    IWL_TRANS_FW_ALIVE  = 1,
}IWL_TRANS_STATE;

/**
 * enum CMD_MODE - how to send the host commands ?
 *
 * @CMD_SYNC: The caller will be stalled until the fw responds to the command
 * @CMD_ASYNC: Return right away and don't wait for the response
 *  response. The caller needs to call iwl_free_resp when done.
 * @CMD_HIGH_PRIO: The command is high priority - it goes to the front of the
 *  command queue, but after other high priority commands. valid only
 *  with CMD_ASYNC.
 * @CMD_SEND_IN_IDLE: The command should be sent even when the trans is idle.
 * @CMD_MAKE_TRANS_IDLE: The command response should mark the trans as idle.
 * @CMD_WAKE_UP_TRANS: The command response should wake up the trans
 *  (i.e. mark it as non-idle).
 */
enum CMD_MODE {
    CMD_SYNC        = 0,
    CMD_ASYNC       = BIT(0),
    CMD_WANT_SKB        = BIT(1),
    CMD_SEND_IN_RFKILL  = BIT(2),
    CMD_HIGH_PRIO       = BIT(3),
    CMD_SEND_IN_IDLE    = BIT(4),
    CMD_MAKE_TRANS_IDLE = BIT(5),
    CMD_WAKE_UP_TRANS   = BIT(6),
};

/**
 * struct iwl_hcmd_dataflag - flag for each one of the chunks of the command
 *
 * @IWL_HCMD_DFL_NOCOPY: By default, the command is copied to the host command's
 *  ring. The transport layer doesn't map the command's buffer to DMA, but
 *  rather copies it to a previously allocated DMA buffer. This flag tells
 *  the transport layer not to copy the command, but to map the existing
 *  buffer (that is passed in) instead. This saves the memcpy and allows
 *  commands that are bigger than the fixed buffer to be submitted.
 *  Note that a TFD entry after a NOCOPY one cannot be a normal copied one.
 * @IWL_HCMD_DFL_DUP: Only valid without NOCOPY, duplicate the memory for this
 *  chunk internally and free it again after the command completes. This
 *  can (currently) be used only once per command.
 *  Note that a TFD entry after a DUP one cannot be a normal copied one.
 */
enum iwl_hcmd_dataflag {
    IWL_HCMD_DFL_NOCOPY = BIT(0),
    IWL_HCMD_DFL_DUP    = BIT(1),
};

/**
 * struct iwl_host_cmd - Host command to the uCode
 *
 * @data: array of chunks that composes the data of the host command
 * @resp_pkt: response packet, if %CMD_WANT_SKB was set
 * @_rx_page_order: (internally used to free response packet)
 * @_rx_page_addr: (internally used to free response packet)
 * @handler_status: return value of the handler of the command
 *  (put in setup_rx_handlers) - valid for SYNC mode only
 * @flags: can be CMD_*
 * @len: array of the lengths of the chunks in data
 * @dataflags: IWL_HCMD_DFL_*
 * @id: id of the host command
 */
typedef struct Iwl_Host_Cmd {
    const void *Data[IWL_MAX_CMD_TBS_PER_TFD];
    IWL_RX_PACKET *Resp_Pkt;
    UINT64 _rx_page_addr;
    UINT32 _rx_page_order;
    INT32 Handler_Status;
    UINT32 Flags;
    UINT16 Length[IWL_MAX_CMD_TBS_PER_TFD];
    UINT8 DataFlags[IWL_MAX_CMD_TBS_PER_TFD];
    UINT8 Id;
}IWL_HOST_CMD;
/**
 * struct iwl_trans_config - transport configuration
 *
 * @op_mode: pointer to the upper layer.
 * @cmd_queue: the index of the command queue.
 *  Must be set before start_fw.
 * @cmd_fifo: the fifo for host commands
 * @no_reclaim_cmds: Some devices erroneously don't set the
 *  SEQ_RX_FRAME bit on some notifications, this is the
 *  list of such notifications to filter. Max length is
 *  %MAX_NO_RECLAIM_CMDS.
 * @n_no_reclaim_cmds: # of commands in list
 * @rx_buf_size_8k: 8 kB RX buffer size needed for A-MSDUs,
 *  if unset 4k will be the RX buffer size
 * @bc_table_dword: set to true if the BC table expects the byte count to be
 *  in DWORD (as opposed to bytes)
 * @queue_watchdog_timeout: time (in ms) after which queues
 *  are considered stuck and will trigger device restart
 * @command_names: array of command names, must be 256 entries
 *  (one for each command); for debugging only
 */
struct iwl_trans_config {
    struct iwl_op_mode *op_mode;

    UINT8 cmd_queue;
    UINT8 cmd_fifo;
    const UINT8 *no_reclaim_cmds;
    UINTN n_no_reclaim_cmds;

    BOOLEAN rx_buf_size_8k;
    BOOLEAN bc_table_dword;
    UINTN queue_watchdog_timeout;
    const CHAR8 *const *command_names;
};

struct iwl_pcie_txq_scratch_buf {
    struct Iwl_Cmd_Header hdr;
    UINT8 buf[8];
    __le32 scratch;
};

VOID
Iwl_Clear_Bit (
    IN IWL_TRANS_CONTEXT *IwlTransContext,
    IN UINT32 reg,
    IN UINT32 mask
    );


extern
INT32
test_bit (
    IN UINT64 nr,
    IN const volatile void *addr
    );


EFI_STATUS
WirelessPciUnMap(
                 IN EFI_PCI_IO_PROTOCOL *PciInstance,
                 IN VOID *Mapping);

EFI_STATUS
WirelessPciMap(
               IN EFI_PCI_IO_PROTOCOL            *PciInstance,
               IN UINT32                          Direction,
               IN VOID                           *HostAddress,
               IN OUT UINTN                      *NumberOfBytes,
               OUT EFI_PHYSICAL_ADDRESS          *DeviceAddress,
               OUT VOID                          **Mapping);

#endif


